# -*- coding: utf-8 -*-
"""
Created on Mon Aug  9 14:51:59 2021

@author: MkZee
"""

import datetime
import functools
import os.path
import sys
import threading
import time
import traceback

import tkinter as tk

from .common import save_dir
from threading import Thread


queue_error_msg = ("A queue.Full error has occured. This is a common error\n"
                   "that has a few possible reasons for being thrown.\n"
                   "Here are solutions in order of most helpful:\n\n"
                   "1) Check for loose cables.\n"
                   "2) Ensure a blocksize of 512 and a buffersize of 40.\n"
                   "3) Increase blocksize(power of 2) and buffersize.\n"
                   "4) Restart your computer and turn off any superfluous\n"
                   "   software while running the test.\n")
                    

def in_thread(thread, wait=True, except_ = None):
    """A function decorator to ensure that a function runs in the given thread.

    Has no effect whenever the function is called from the same thread it is set
    to be in.

    thread: one of 'GuiThread', 'MainThread'
        the thread to run the function in

    wait : bool
        whether to wait for the function to return in the other thread.
        if this is False, it may return None if called from a different thread.

    except_:
        an exception or tuple of exceptions which will be passed to the caller
        if raised by the function. (as if used in an except clause).
        Has no effect if wait is False

    Example:

        @in_thread('MainThread')
        def my_mainthread_func(*args, **kwargs):
            print('Hello from the main thread!')

    """

    exceptions = except_

    # convert exceptions to catch nothing if they are not applicable
    if not wait or exceptions is None:
        exceptions = tuple()

    if thread not in ('MainThread', 'GuiThread'):
        raise ValueError(f'Thread {thread} does not exist')

    def decorator(func):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):

            if thread == threading.current_thread().getName():
                # we are already in that thread!!
                # run the function as normal
                return func(*args, **kwargs)

            switch_obj = _dec_return(func, args, kwargs, exceptions)

            if thread == 'MainThread':
                main.callback(switch_obj.callbacker)
            elif thread == 'GuiThread':
                main.gui_thread.callback(switch_obj.callbacker)

            if wait:
                # wait until function is finished if applicable
                while not switch_obj.finished:
                    if thread == 'MainThread':
                        # keep the gui responsive
                        main.win.update_idletasks()

                    time.sleep(0.1)

                if switch_obj.exc is not None:
                    raise switch_obj.exc

                return switch_obj.return_val

        return wrapper
    return decorator


class _dec_return:
    def __init__(self, func, args, kwargs, exceptions):

        self.func = func
        self.finished = False
        self.args = args
        self.kwargs = kwargs

        self.exc = None
        self.return_val = None
        self.exceptions = exceptions

    def callbacker(self):
        try:
            self.return_val = self.func(*self.args, **self.kwargs)

        except self.exceptions as e:
            self.exc = e

        except BaseException as e:
            show_error(e)

        self.finished = True


class GuiThread(Thread):

    def callback(self, function):
        """
        Calls a function in the GUIThread

        Parameters
        ----------
        function : callable

        """
        
        self._callbacks.insert(0, function)

    def __init__(self, win_class=tk.Tk):
        super().__init__()

        self.setName('GuiThread')

        self.setDaemon(True)
        self._callbacks = []
        self.win_class = win_class
        self.win = None

    def run(self):
        # construct window
        self.win = self.win_class()

        # register callback for error handling in gui-thread
        self.win.report_callback_exception = lambda typ, exc, trace: show_error(exc)

        self.win.after(100, self._main_loop_ext)
        self.win.mainloop()

    def _main_loop_ext(self):

        try:
            # grab a function
            f = self._callbacks.pop()
        except IndexError:pass
        else:
            try:
                f()
            except Exception as e:
                # don't crash in case of error
                show_error(e)

        # run this function again
        self.win.after(100, self._main_loop_ext)

    @in_thread('GuiThread')
    def stop(self):
        if self.win is not None:
            self.win.destroy()


class Main():
    def __init__(self, win_class=tk.Tk):
        global main
        main = self
        self._break = False
        self._callbacks = []
        self.is_running = False
        self.last_error = None

        # constructs gui in new thread
        self.gui_thread = GuiThread(win_class)
        self.gui_thread.start()

        while self.gui_thread.win is None:
            time.sleep(0.01)

        self.win = self.gui_thread.win

    def callback(self, function):
        """
        Calls a function in the main thread.

        Parameters
        ----------
        function : callable

        Returns
        -------
        None.

        """
        
        self._callbacks.insert(0, function)

    def main_loop(self):
        while True:
            if self._break:
                self.is_running = False
                self.gui_thread.join()
                return
            try:
                try:
                    f = self._callbacks.pop()
                except IndexError:
                    # if no callbacks to be called
                    self.is_running = False
                    time.sleep(0.2)
                else:
                    self.is_running = True
                    f()

            except Abort_by_User:pass

            except SystemExit:
                break

            except KeyboardInterrupt:pass

            except Exception as e:

                # prints exception without exiting main thread
                show_error(e)

        # thread is ending
        self.is_running = False

        self.gui_thread.join()
        return

    def stop(self):
        self._break = True


class SingletonWindow(type):
    """
    A meta class to ensure one instance of window classes.

    This class keeps track of each class that uses it as a metaclass and will
    return the existing instance of a class that has already been created.

    Examples
    --------

    Create a new window class that can only be opened once
    >>>class MyWin(tk.Toplevel, metaclass = SingletonWindow):
    >>>    pass
    """
    
    # dict to hold existing class instances
    _instances = {}

    def __call__(cls, *args, **kwargs):
        # check if class is in list and still exists
        if cls not in cls._instances or not cls._instances[cls].winfo_exists():
            # create new instance of the class
            cls._instances[cls] = super(SingletonWindow, cls).__call__(*args, **kwargs)
        else:
            # already exists, focus window
            cls._instances[cls].focus_force()
            cls._instances[cls].bell()
        # Return class instance, either old or newly created
        return cls._instances[cls]

@in_thread('GuiThread')
def _show_error(err_name, msg):
    tk.messagebox.showerror(err_name, msg)

def show_error(exc=None, err_func=None):

    err_time = datetime.datetime.now()

    # dir for traceback.log file
    base_fold = save_dir
    # full path to log file
    err_log = os.path.join(base_fold,'traceback.log')

    try:
        # get traceback info in a string
        tb_str = traceback.format_exc()
        # make sure directory exists
        os.makedirs(base_fold, exist_ok=True)

        with open(err_log, 'at') as f:
            # write Heading for start of traceback
            f.write(f'##Error Encountered on {err_time.strftime("%c")}:\n')
            # write traceback
            f.write(tb_str)
            # add an extra newline for seperation
            f.write('\n')
    except (OSError, IOError):
        print('Unable to write traceback.log')

    traceback.print_exc()

    if exc is None:
        exc = sys.exc_info()[1]
    if isinstance(exc, tuple):
        exc = exc[1]

    if not exc:
        # no error
        return

    if isinstance(exc, InvalidParameter) and err_func is None:
        main.win.show_invalid_parameter(exc)
    else:
        err_name, msg = format_error(exc)

        if err_func:
            err_func(err_name, msg)
        elif err_name == "queue.Full":
            _show_error(err_name, queue_error_msg)
        else:
            _show_error(err_name, msg)

def format_error(exc):
    msg = str(exc)
    err_name = exc.__class__.__name__

    if not msg and isinstance(exc, Exception):
        if 'error' in err_name.lower():
            descriptor = ''
        else:
            descriptor = ' error'
        if err_name[0].lower() in 'aeiou':
            article = 'An'
        else:
            article = 'A'

        msg = f'{article} "{err_name}"{descriptor} occurred.'

        err_name = 'Error'
    elif isinstance(exc, Abort_by_User):
        err_name = 'Measurement Stopped'
        msg = 'Aborted by user.'

    return err_name, msg


class InvalidParameter(ValueError):
    """Raised when a user fails to input a required parameter"""

    def __init__(self, parameter, message=None, *args, **kwargs):

        super().__init__(f'{parameter}: {message}', *args, **kwargs)

        self.parameter = parameter
        self.message = message


class Abort_by_User(BaseException):
    """
    Raised when user presses 'Abort test'

    Inherits from BaseException because it is not an error and therefore
    won't be treated as such.
    """
    
    def __init__(self, *args, **kwargs):
        super().__init__('Measurement aborted by user', *args, **kwargs)



from dash import dcc
from dash import html
from dash.dependencies import Input, Output, State

import base64
import importlib
import io
import json
import numpy as np
import os
import pandas as pd
import plotly.graph_objects as go
import re

from mcvqoe.hub.eval_app import app
from mcvqoe.hub.common import save_dir as default_data_dir

import mcvqoe.mouth2ear
import mcvqoe.intelligibility
import mcvqoe.psud
import mcvqoe.accesstime
import mcvqoe.tvo


# --------------[Measurement Globals]-----------------------------------

measurements = [
    'm2e',
    'intell',
    'psud',
    'access',
    'tvo',
    'diagnostics',
    ]

# --------------[General Style]--------------------------------------------

plotly_default_color = '#edeef0'
style_links = {
    'width': '240px',
    # 'height': '60px',
    'display': 'block',
    # 'textAlign': 'center',
    'overflow-wrap': 'break-work',
    'lineHeight': '60px',
    'borderWidth': '1px',
    'borderStyle': 'solid',
    'borderRadius': '5px',
    'margin': '10px',
    'padding': '10px',
    'backgroundColor': '#E5ECF6',
        }

# --------------[Top of Page information]-----------------------

def mcv_headers(measurement):
    if measurement == 'm2e':
        full_meas = 'Mouth-to-ear latency'
    elif measurement == 'psud':
        full_meas = 'Probability of successful delivery'
    elif measurement == 'intell':
        full_meas = 'Speech intelligibility'
    elif measurement == 'access':
        full_meas = 'Access delay'
    elif measurement == 'tvo':
        full_meas = 'Transmit volume optimization'
    elif measurement == 'diagnostics':
        full_meas = 'Measurement diagnostics'
    else:
        full_meas = 'Undefined measurement'
            
    # TODO: Add other measurements
    children = [
        html.H1('MCV QoE Measurements Processing'),
        html.H3(f'{full_meas} data analysis')
        ]
    return children

#---------------[Parsing Data]----------------------------

def parse_contents(contents, filename):
    """
    Parse contents of uploaded data

    Parameters
    ----------
    contents : TYPE
        DESCRIPTION.
    filename : TYPE
        DESCRIPTION.

    Returns
    -------
    children : TYPE
        HTML representations of filename or error string.
    df : pd.DataFrame
        Data stored as dataframe.

    """
    
    content_type, content_string = contents.split(',')
    decoded = base64.b64decode(content_string)
    fname, ext = os.path.splitext(filename)

    try:
        if ext == '.csv':
            # Load in data in fpath
            try:
                df = pd.read_csv(io.StringIO(decoded.decode('utf-8')))
            except pd.errors.ParserError:
                # If parsing failed, access data, skip 2 rows (TVO)
                df = pd.read_csv(io.StringIO(decoded.decode('utf-8')), skiprows=2)
                if len(df.columns) == 1:
                    # If everything got squished into one column it is access
                    df = pd.read_csv(io.StringIO(decoded.decode('utf-8')), skiprows=3)
                else:
                    # Actually is TVO, store opt
                    df_opt = pd.read_csv(io.StringIO(decoded.decode('utf-8')), nrows=1)
                    df = (df, df_opt)
    except Exception as e:
        print(e)
        children =  html.Div([
            'There was an error processing this file'
            ])
        return children, None
    
    children = format_data_filename(filename)

    return children, df

def load_json_data(jsonified_data, measurement):
    """
    Load json data as evaluate object for given measurement

    Parameters
    ----------
    jsonified_data : str
        Jsonified dict of jsonified dataframes. See evaluate.to_json() for any
        measurement evaluate class.
    
    measurement : str
        One of psud, access, m2e, intell

    Returns
    -------
    eval_obj : mcvqoe.mouth2ear.evaluate, mcvqoe.accesstime.evaluate, mcvqoe.psud.evaluate, mcvqoe.intelligibility.evaluate
        Evaluate object of all stored data.

    """

    modules = {'access': 'mcvqoe.accesstime',
               'intell': 'mcvqoe.intelligibility',
               'm2e': 'mcvqoe.mouth2ear',
               'psud': 'mcvqoe.psud',
               'tvo': 'mcvqoe.tvo',
               'diagnostics': 'mcvqoe.diagnostics',
               }
    eval_obj = importlib.import_module(modules[measurement]).evaluate(json_data=jsonified_data)
    # eval_obj = eval(modules[measurement]).evaluate(json_data=jsonified_data)
        
    return eval_obj

# TODO: Make these leverage evaluate classes better, rather than redoing work...
def find_cutpoints(fname, df, measurement):
    '''
    Try to find cutpoints based on fname.
    
    Looks for cutpoints in:
        * mcvqoe.hub.common.save_dir as default_data_dir
        * sees if it can find matching cutpoints from measurement default/included audio
    If cannot find it in either needs to do something smart to inform user of limitation
      * Say open data from measurement gui/make sure that they have stuff stored appropriately (Reference?)
    '''
    
    data_dir = os.path.join(default_data_dir, measurement, 'data')
    # Identify session identification string
    session_pattern = re.compile(r'(capture_.+_\d{2}-\w{3}-\d{4}_\d{2}-\d{2}-\d{2})(?:.*\.csv)')
    session_search = session_pattern.search(fname)
    if session_search is not None:
        session_id = session_search.groups()[0]
    else:
        # TODO: Be smarter than this
        raise RuntimeError('No valid session id found in uploaded data')
   
    # Construct wav folder path based off capture id
    wav_dir = os.path.join(data_dir, 'wav', session_id)
    # Determine if wav directory exists
    if os.path.exists(wav_dir):        
        # Initialize cutpoints dictionary
        cps = dict()
        # TODO: This logic only works for PSuD, extract from access data file
        for file in np.unique(df['Filename']):
            # Construct path to cutpoint file
            cp_name = f'Tx_{file}.csv'
            cp_path = os.path.join(wav_dir, cp_name)
            
            # Load cutpoints
            cp = pd.read_csv(cp_path)
            
            # Store as json in dict
            cps[file] = cp.to_json()
    else:
        # Try to find the cutpoints from the defaults included in the package
        if measurement == 'psud':
            # Get default audio sets and audio path
            [default_audio_sets, default_audio_path] = mcvqoe.psud.measure.included_audio_sets()
            # Initialize cutpoints dictionary
            cps = dict()
            
            # Search string extract set number from filenames
            set_search = re.compile(r'(?:[FM]\d_n\d+_s)(\d+)(?:_c\d+)')
            for file in np.unique(df['Filename']):
                # Search for set
                search_res = set_search.search(file)
                if search_res is not None:
                    aset = 'set' + search_res.groups()[0]
                    # Make sure it is in default audio sets
                    if aset in default_audio_sets:
                        # Construct path to cutpoint file
                        cp_name = f'{file}.csv'
                        cp_path = os.path.join(default_audio_path, aset, cp_name)
                        
                        # Load cutpoints
                        cp = pd.read_csv(cp_path)
                        
                        # Store in json dict
                        cps[file] = cp.to_json()
                    else:
                        # TODO: Figure out how to flag this for users
                        raise RuntimeError('Cannot find cutpoints, say something to user')
                else:
                    # TODO: Figure out how to flag this for users
                    raise RuntimeError('Cannot find cutpoints, say something to user')
        elif measurement == 'access':
            # Get default audio sets and audio path
            default_audio_path = mcvqoe.accesstime.measure.included_audio_path()
            default_files = os.listdir(default_audio_path)
            
            # Extract talker word combo from file name
            tw_search_str = r'([FM]\d)(_b\d{1,2}_w\d_)(\w+)(?:.csv)'
            tw_search_var = re.compile(tw_search_str)
            tw_search = tw_search_var.search(fname)
            if tw_search is None:
                raise RuntimeError(f'Unable to determine talker word from filename {fname}')
            
            # Get talker word identifier for cutpoitns
            talker_word_name = tw_search.group()
            
            # Get talker word combo for storing
            talker, bw_index, word = tw_search.groups()
            talker_word = talker + ' ' +  word
            
            # Get matching cutpoitn file name from defaults
            cp_name = [x for x in default_files if talker_word_name in x]
            if len(cp_name) == 1:
                cp_name = cp_name[0]
            else:
                # TODO: Figure out how to flag this for users
                raise RuntimeError('Cannot find cuptpoints, say something to user')
            cp_path = os.path.join(default_audio_path, cp_name)
            
            # Load cutpoints file
            cp = pd.read_csv(cp_path)
            
            # Save in dict as json
            cps = {talker_word: cp.to_json()}
            
            # Extract session id from file name
            name = fname.replace('_' + talker_word_name, '')
            # TODO: fs hardcoded...not ideal
            fs = 48000
            
            # Determine length of T from cutpoints
            T = cp.loc[0]['End']/fs
            
            # Store PTT time relative to start of P1 
            df['time_to_P1'] = T - df['PTT_time']
            # Store session name
            df['name'] = name
            # Store talker word combo
            df['talker_word'] = talker_word
        # Store all measurement data and cutpoints for this session file
    
    out_json = {
        'measurement': df.to_json(),
        'cutpoints': cps,
        }
    return out_json


for measurement in measurements:
    @app.callback(
        Output(f'{measurement}-output-data-upload', 'children'),
        Output(f'{measurement}-json-data', 'data'),
        Output(f'{measurement}-initial-data-passed', 'children'),
        Input(f'{measurement}-upload-data', 'contents'),
        Input(f'{measurement}-upload-data', 'filename'),
        State(f'{measurement}-initial-data-passed', 'children'),
        State(f'{measurement}-json-data', 'data'),
        )
    
    def update_output(list_of_contents, list_of_names,
                      initial_data_flag, initial_data, measurement=measurement):
        """
        Process uploaded data and store csv files as json
    
        Parameters
        ----------
        list_of_contents : TYPE
            DESCRIPTION.
        list_of_names : TYPE
            DESCRIPTION.
        list_of_dates : TYPE
            DESCRIPTION.
    
        Returns
        -------
        children : TYPE
            DESCRIPTION.
        final_json : TYPE
            DESCRIPTION.
    
        """
        
        # TODO: Add comments for this stuff
        if initial_data_flag == 'True':
            final_json = initial_data
            test_dict = json.loads(final_json)
            children = []
            if 'error' in test_dict:
                filenames = test_dict['test_info']

            else:    
                filenames = test_dict['test_info'].keys()
                
            for filename in filenames:
                children.append(format_data_filename(filename))
        else:
            # TODO: Figure out what to do with cutpoints here...
            if list_of_contents is not None:
                try:
                    children = []
                    # TODO: Figure out a better variable name/place to store this
                    out_json = {}
                    test_info = {}
                    
                    if measurement == 'tvo' and len(list_of_names) > 1:
                        raise ValueError('Cannot process more than one TVO csv at a time')
                    for c, filename in zip(list_of_contents, list_of_names):
                        child, df = parse_contents(c, filename)
                        
                        if isinstance(df, tuple) and len(df) == 2:
                            # Grab extra dataframe if there (right now just for TVO)
                            df_extra = df[1]
                            df = df[0]
                            
                        df['name'] = filename

                        children.append(child)
                        
                        if measurement == 'psud':
                            out_json[filename] = find_cutpoints(filename, df, measurement)
                            
                        elif measurement == 'access':
                            out_json[filename] = find_cutpoints(filename, df, measurement)
                        
                        elif measurement == 'tvo':
                            out_json[filename] = {
                                'measurement': df.to_json(),
                                'optimal': df_extra.to_json(),
                                }
                            
                        else:
                            out_json[filename] =  {
                                'measurement': df.to_json()
                                }
                        test_info[filename] = 'uploaded-data'
                        
                    # Initialize dataframe for storing all data
                    data_df = []
                    # Initialize cutpoints dict
                    cps = dict()
                    optimal_list = []
                    for sesh, sesh_dict in out_json.items():
                        # Load data frame
                        df = pd.read_json(sesh_dict['measurement'])
                        # Add to main dataframe
                        data_df.append(df)
                        if 'cutpoints' in sesh_dict:
                            cps[sesh] = sesh_dict['cutpoints']

                        if 'optimal' in sesh_dict:
                            optimal_list.append(pd.read_json(sesh_dict['optimal']))
                    
                    if len(optimal_list) > 0:
                        optimal = pd.concat(optimal_list)
                        optimal = optimal.to_json()
                    else:
                        optimal = None
                    
                    data_df = pd.concat(data_df, ignore_index=True)

                    # Store in dictionary used by mcvqoe.accesstime.load_json()
                    prep_json = {
                        'measurement': data_df.to_json(),
                        'cps': cps,
                        'test_info': test_info,
                        'optimal': optimal,
                            }
                    # Dump dict to json
                    
                    final_json = json.dumps(prep_json)

                except Exception as e:
                    out_info = {'test_info': list_of_names,
                            'error': e.args,
                        }
                    print(f'e.args: {e.args}\ntest_info: {list_of_names}')
                    final_json = json.dumps(out_info)
                    
            else:
                children = None
                final_json = None
        initial_data_flag = html.Div('False')

        return children, final_json, initial_data_flag
    
# --------------[Data Filename Formattting]-----------------------

style_data_filename = {
            'fontSize': 12,
            }

def format_data_filename(filename):
    children = html.Div([
        html.P([filename])
        ],
        style=style_data_filename)
    return children

#-------------------[Results Styles]-------------------------------------

style_results = {
    'backgroundColor': plotly_default_color,
    'width': '100%',
    'borderWidth': '1px',
    'borderStyle': 'solid',
    'borderRadius': '5px',
    'textAlign': 'left',
    'margin': '10px',
    'padding': '10px',
    }
style_measurement_format = {
    'fontSize': 12,
    'backgroundColor': plotly_default_color,
    'width': '50%',
    'borderWidth': '1px',
    'borderStyle': 'solid',
    'borderRadius': '5px',
    'marginTop': '10px',
    'marginRight': '10px',
    'padding': '10px',
    }

style_result_filters_dropdown = {
    'width': '30%',
    'display': 'inline-block'
    }

def measurement_results_filters(measurement):
    if measurement == 'psud':
        raw_intell_options = np.arange(0.5, 1.01, 0.1)
        intell_options = [{'label': np.round(x, 1), 'value': np.round(x, 1)} for x in raw_intell_options]
        default_intell_options=[0.5,]
        
        raw_msg_len_options = np.arange(1, 11, 1)
        msg_len_options = [{'label': x, 'value': x} for x in raw_msg_len_options]
        default_msg_len = [1, 3, 5, 10]
        
        children = html.Div([
            html.Div([
                html.Label('Method'),
                dcc.Dropdown(
                    id=f'{measurement}-method',
                    options=[
                        {'label': 'Every Word Critical (EWC)', 'value': 'EWC'},
                        {'label': 'Average Message Intelligibility (AMI)', 'value': 'AMI'},
                        ],
                    multi=True,
                    value=['EWC', 'AMI'],
                    ),
                ],
                style=style_result_filters_dropdown,
                ),
            html.Div([
                html.Label('Intelligibility Threshold'),
                dcc.Dropdown(
                    id=f'{measurement}-intelligibility-threshold',
                    options=intell_options,
                    multi=True,
                    value=default_intell_options,
                    
                    ),
                ],
                style=style_result_filters_dropdown,
                ),
            html.Div([
                html.Label('Message Length'),
                dcc.Dropdown(
                    id=f'{measurement}-message-length',
                    options=msg_len_options,
                    multi=True,
                    value=default_msg_len,
                    ),
                ],
                style=style_result_filters_dropdown,
                )
            ])
    elif measurement == 'access':
        raw_alpha_options = np.arange(0.5, 1, 0.01)
        alpha_options = [{'label': np.round(x, 2), 'value': np.round(x, 2)} for x in raw_alpha_options]
        default_alpha_options = [0.9,]
        children = html.Div([
            html.Div([
                html.Label('Alpha'),
                dcc.Dropdown(
                    id=f'{measurement}-alpha',
                    options=alpha_options,
                    multi=True,
                    value=default_alpha_options,
                    ),
                ],
                style=style_result_filters_dropdown,
                )
            ])
    else:
        children = None
    return children

digit_range = [1, 6]
digit_default = 4

def measurement_digits(display, digits=digit_default, measurement=''):
    style = style_measurement_format
    
    style['display'] = display
    children = html.Div([
        html.Label(f'Number of Digits in Results ({digit_range[0]} - {digit_range[1]}):'),
        dcc.Input(id=f'{measurement}-measurement-digits',
                  value=digits,
                  type='number',
                  min=digit_range[0],
                  max=digit_range[1],
                  style={
                      'backgroundColor': '#f0f2f5',
                      # 'width': '100%',
                      }
                  ),
        ],
        style=style_measurement_format,
        )
    return children

def pretty_numbers(x, digits=digit_default):
    if isinstance(x, list):
        pretty_vals = []
        for xv in x:
            pretty_vals.append(pretty_numbers(xv, digits=digits))
    else:
        pretty_vals = np.round(x, digits)
    return pretty_vals

#-------------------[Filter Styles]----------------------------------------

radio_button_style = {'width': '15%', 'display': 'inline-block'}
radio_labels_style = {'display': 'inline-block'}
dropdown_style = {'width': '45%', 'display': 'inline-block'}

def dropdown_filters(measurement):
    if measurement == 'access':
        children = [html.Div([
                    html.Label('Talker Select'),
                    dcc.Dropdown(
                        id=f'{measurement}-talker-select',
                        multi=True,
                        )
                    ], style=dropdown_style),
                ]
    elif measurement == 'diagnostics':
        children = []
    elif measurement in measurements:
        children = [html.Div([
                    html.Label('Session Select'),
                    dcc.Dropdown(
                        id=f'{measurement}-session-select',
                        multi=True
                        )
                    ], style=dropdown_style),
                html.Div([
                    html.Label('Talker Select'),
                    dcc.Dropdown(
                        id=f'{measurement}-talker-select',
                        multi=True,
                        )
                    ], style=dropdown_style),
                ]
    else:
        children = [html.Div('Undefined Measurement')]
    return children

def radio_filters(measurement):
    if measurement == 'm2e':
        children = [
            html.Div([
                html.Label('Data thinning'),
                dcc.RadioItems(
                    id=f'{measurement}-thin-select',
                    options=[{'label': 'Thinned', 'value': 'True'},
                             {'label': 'All', 'value': 'False'}
                             ],
                             value='True',
                             labelStyle=radio_labels_style,
                    ),
                ], style=radio_button_style),
            html.Div([
                html.Label('X-axis'),
                dcc.RadioItems(
                    id=f'{measurement}-x-axis',
                    options = [{'label': 'Trial', 'value': 'index'},
                               {'label': 'Timestamp', 'value': 'Timestamp'},
                               ],
                    value='index',
                    labelStyle=radio_labels_style,
                    ),
                ],
                style=radio_button_style
                ),
            ]
    elif measurement == 'intell':
        children = [
            # TODO: Generalize this
            html.Div([
                html.Label('X-axis'),
                dcc.RadioItems(
                    id=f'{measurement}-x-axis',
                    options = [{'label': 'Trial', 'value': 'index'},
                               {'label': 'Timestamp', 'value': 'Timestamp'},
                               ],
                    value='index',
                    labelStyle=radio_labels_style,
                    ),
                ],
                style=radio_button_style
                ),
            ]
    elif measurement == 'psud':
        children = [
            # TODO: Generalize this
            html.Div([
                html.Label('Intelligibility'),
                dcc.RadioItems(
                    id=f'{measurement}-intell-type',
                    options = [
                        {'label': 'Message', 'value': 'message'},
                        {'label': 'Word', 'value': 'word'},
                        ],
                    value='message',
                    labelStyle=radio_labels_style,
                    ),
                ],
                style=radio_button_style,
                ),
            html.Div([
                html.Label('X-axis'),
                dcc.RadioItems(
                    id=f'{measurement}-x-axis',
                    options = [{'label': 'Trial', 'value': 'index'},
                               {'label': 'Timestamp', 'value': 'Timestamp'},
                               ],
                    value='index',
                    labelStyle=radio_labels_style,
                    ),
                ],
                style=radio_button_style
                ),
            ]
    elif measurement == 'access':
        children = [
            # TODO: Generalize this
            html.Div([
                html.Label('Intelligibility'),
                dcc.RadioItems(
                    id=f'{measurement}-intell-type',
                    options = [
                        {'label': 'Relative to asymptotic', 'value': 'relative'},
                        {'label': 'Raw Intelligibility', 'value': 'intelligibility'},
                        ],
                    value='relative',
                    labelStyle=radio_labels_style,
                    ),
                ],
                style=radio_button_style,
                ),
            html.Div([
                html.Label('Plot raw data points'),
                dcc.RadioItems(
                    id=f'{measurement}-show-raw',
                    options = [{'label': 'True', 'value': 'True'},
                               {'label': 'False', 'value': 'False'},
                               ],
                    value='True',
                    labelStyle=radio_labels_style,
                    ),
                ],
                style=radio_button_style
                ),
            html.Div([
                html.Label('Use start-of-word correction (recommended)'),
                dcc.RadioItems(
                    id=f'{measurement}-sowc',
                    options = [{'label': 'True', 'value': 'True'},
                               {'label': 'False', 'value': 'False'},
                               ],
                    value='True',
                    labelStyle=radio_labels_style,
                    ),
                ],
                style=radio_button_style
                ),
            ]
    elif measurement == 'tvo':
        children = [
            html.Div([
                html.Label('X-axis'),
                dcc.RadioItems(
                    id=f'{measurement}-x-axis',
                    options = [{'label': 'Volume', 'value': 'Volume'},
                               {'label': 'Trial', 'value': 'index'},
                               {'label': 'Timestamp', 'value': 'Timestamp'},
                               ],
                    value='Volume',
                    labelStyle=radio_labels_style,
                    ),
                ],
                style=radio_button_style
                ),
            ]
    elif measurement == 'diagnostics':
        children = []
    else:
         children = [html.Div('Undefined measurement')]   
    return children

#------------------------[Figure and Plots]----------------------------

def blank_fig():
    """Make a blank plotly figure for a placeholder prior to data load"""
    
    fig = go.Figure(go.Scatter(x=[], y = []))
    fig.update_layout(template = None)
    fig.update_xaxes(showgrid = False, showticklabels = False, zeroline=False)
    fig.update_yaxes(showgrid = False, showticklabels = False, zeroline=False)
    
    return fig

# This color palette was defined to be both colorblind-friendly and meet WCAG AA
# contrast standards. 
# See:
# https://davidmathlogic.com/colorblind/#%23E31A1C-%235E3C99-%23018E79-%23948A00-%23004E79-%23823900-%23000000
# https://webaim.org/resources/contrastchecker/
plotly_color_palette = [
    '#E31A1C', # Red
    '#5E3C99', # Plum
    '#018E79', # Blue/green
    '#948A00', # Gold/sand
    '#004E79', # Dark Blue
    '#823900', # Dark orange/red
    '#000000', # Black
    ]
    

def measurement_plots(measurement):
    
    if measurement == 'm2e':
        children = [
            # ------------[Scatter Plot]---------------------
            html.Div([
                dcc.Graph(id=f'{measurement}-scatter',
                          figure=blank_fig(),
                  ),
                ], className='twelve columns'),
            # ----------------[Histogram]---------------------
            html.Div([
                dcc.Graph(id=f'{measurement}-hist',
                          figure=blank_fig(),
                  ),
                ], className='six columns'),
            ]
    elif measurement == 'intell':
        children = [
            # -----------------[Bar Graph]------------------
            html.Div([
                dcc.Graph(id=f'{measurement}-bar',
                          figure=blank_fig(),
                          ),
                ], style={'text-align': 'center'}),
            # ----------------[Scatter Plot]----------------
            html.Div([
                dcc.Graph(id=f'{measurement}-hist',
                          figure=blank_fig(),
                  ),
                ], className='twelve columns'),
            # ----------------[Histogram]---------------------
            html.Div([
                dcc.Graph(id=f'{measurement}-scatter',
                          figure=blank_fig(),
                  ),
                ], className='twelve columns'),
            ]
    elif measurement == 'psud':
        children = [
            # --------------[PSuD vs msg length Plot] ---------
            html.Div([
                dcc.Graph(id=f'{measurement}-plot',
                          figure=blank_fig(),
                          ),
                ], className='twelve columns'),
            # --------------[Intell Scatter Plot]------------
            html.Div([
                dcc.Graph(id=f'{measurement}-scatter',
                          figure=blank_fig(),
                          ),
                ], className='twelve columns'),
            # ----------------[Test chain histogram]-------------
            html.Div([
                dcc.Graph(id=f'{measurement}-hist',
                          figure=blank_fig(),
                          ),
                ], className='twelve columns'),
            ]
    elif measurement == 'access':
        children = [
            # --------------[Access plot] ---------
            html.Div([
                dcc.Graph(id=f'{measurement}-plot',
                          figure=blank_fig(),
                          ),
                ], className='twelve columns'),
            # --------------[Intell Scatter Plot]------------
            html.Div([
                dcc.Graph(id=f'{measurement}-intell',
                          figure=blank_fig(),
                          ),
                ], className='twelve columns'),
            ]
    elif measurement == 'tvo':
        children = [
            html.Div([
                dcc.Graph(id=f'{measurement}-plot',
                          figure=blank_fig(),
                          ),
                ], className='twelve columns'),
            ]
    elif measurement == 'diagnostics':
        children = [
            #-------------[A-weight plot]-----------
            html.Div([
                dcc.Graph(id=f'{measurement}-aweight',
                          figure=blank_fig(),
                          ),
                ], className='twelve columns'),
            # --------------[Peak dBfs Plot]------------
            html.Div([
                dcc.Graph(id=f'{measurement}-peak',
                          figure=blank_fig(),
                          ),
                ], className='twelve columns'),
            #-------------[FSF plot]--------------------
            html.Div([
                dcc.Graph(id=f'{measurement}-fsf',
                          figure=blank_fig(),
                          ),
                ], className='twelve columns'),
            ]
            
    else:
        
        children = [html.Div('Undefined measurement')]   
    return children

#-------------[Create Layout Template]-------------------------

def layout_template(measurement):
    
    layout = html.Div([
        # TODO: This should probably be local or session
        # Element to store json representations of data
        dcc.Store(id=f'{measurement}-json-data',
                  storage_type='memory',
                  ),
        html.Div(
            mcv_headers(measurement),
            id=f'{measurement}-headers',),
        #--------------[Upload Data]------------------------------
        dcc.Upload(
            id=f'{measurement}-upload-data',
            children=html.Div([
                'Drag and drop or ',
                html.A('Select Files')
                ]),
            style={
                'width': '50%',
                'height': '60px',
                'lineHeight': '60px',
                'borderWidth': '1px',
                'borderStyle': 'dashed',
                'borderRadius': '5px',
                'textAlign': 'center',
                'margin': '10px'
                },
            multiple=True
            ),
        html.Div(id=f'{measurement}-output-data-upload'),
        html.Div('False',
                 id=f'{measurement}-initial-data-passed',
                 style={'display': 'none'}),
        html.Hr(),
        html.Div(
            measurement_results_filters(measurement),
            id=f'{measurement}-results-filters', # Unused for m2e and intell, only for psud and access
            className='twelve columns',
            ),
        html.Div([
            html.Div(
                id=f'{measurement}-measurement-results',
                className='eight columns',
                ),
            html.Div(
                measurement_digits(display='none', measurement=measurement),
                id=f'{measurement}-measurement-formatting',
                className='four columns',
                ),
            ]),
        html.H3('Plotting Dashboard',
                # style={
                #     'topMargin': '10px',
                #     },
                className='twelve columns'),
        # html.Br(),
        # html.Br(),
        # html.Br(),
        # ----------------[Dropdowns]------------------
        html.Div(
            dropdown_filters(measurement),
            id=f'{measurement}-filters-dropdown',
            style={
                'marginTop': '10px',
                },
            ),
        # --------------------[Radio Button Filtering]--------------------------
        html.Div(
            radio_filters(measurement),
            id=f'{measurement}-filters-buttons',
            ),
        # -------------------------[Plots]----------------------------------------
        html.Div(
            measurement_plots(measurement),
            id=f'{measurement}-plots',
            ),
        # ----------------------[Bottom Navigation]-------------------------
        html.Br(),
        html.Div([
            html.Hr(),
            dcc.Link('Return to measurement selection', href='/measurement_select',
                     style=style_links,
                )
            ],
            className='twelve columns',
            
            ),
        ])
    return layout

def failed_process(measurement, msg=('', )):
    
    if measurement == 'access':
        none_dropdown = [{'label': 'N/A', 'value': 'None'}]
        
        default_msg = ('Access object could not be processed.\n', )
        if msg[0] == 'Optimal parameters not found: Number of calls to function has reached maxfev = 600.':
            debug_help = ('This measurement is invalid. Try rerunning the test with a smaller Time Increase per Step to increase resolution.', )
        else:
            debug_help = ('', )
        msg = default_msg + msg + debug_help
        res = html.Div(msg)
        res_formatting = measurement_digits('none',
                                            measurement=measurement)
        fig_plot = blank_fig()
        fig_intell = blank_fig()
        talker_options = none_dropdown
        
        return_vals = (
                res,
                res_formatting,
                fig_plot,
                fig_intell,
                talker_options,
                )
    elif measurement == 'm2e':
        none_dropdown = [{'label': 'N/A', 'value': 'None'}]
        default_msg = ('Mouth-to-ear latency object could not be processed.\n', )
        debug_help = ('', )
        msg = default_msg + msg + debug_help
        res = html.Div(msg)
        res_formatting = measurement_digits('none',
                                            measurement=measurement)
        fig_scatter = blank_fig()
        fig_histogram = blank_fig()
        talker_options = none_dropdown
        session_options = none_dropdown
            # )
        return_vals = (
                res,
                res_formatting,
                fig_scatter,
                fig_histogram,
                talker_options,
                session_options
                )
    elif measurement == 'psud':
        none_dropdown = [{'label': 'N/A', 'value': 'None'}]
        # return_vals = (
        default_msg = ('PSuD object could not be processed.\n', )
        debug_help = ('', )
        msg = default_msg + msg + debug_help
        res = html.Div(msg)
        res_formatting = measurement_digits('none',
                                            measurement=measurement)
        fig_plot = blank_fig()
        fig_scatter = blank_fig()
        fig_histogram = blank_fig()
        talker_options = none_dropdown
        session_options = none_dropdown
            # )
        return_vals = (
                res,
                res_formatting,
                fig_plot,
                fig_scatter,
                fig_histogram,
                talker_options,
                session_options
                )
    elif measurement == 'intell':
        none_dropdown = [{'label': 'N/A', 'value': 'None'}]
        # return_vals = (
        default_msg = ('Intelligibility object could not be processed.\n', )
        debug_help = ('', )
        msg = default_msg + msg + debug_help
        res = html.Div(msg)
        res_formatting = measurement_digits('none',
                                            measurement=measurement)
        fig_scatter = blank_fig()
        fig_histogram = blank_fig()
        fig_bar = blank_fig()
        talker_options = none_dropdown
        session_options = none_dropdown
            # )
        return_vals = (
                res,
                res_formatting,
                fig_scatter,
                fig_histogram,
                fig_bar,
                talker_options,
                session_options
                )
    elif measurement == 'tvo':
        none_dropdown = [{'label': 'N/A', 'value': 'None'}]
        # return_vals = (
        default_msg = ('TVO object could not be processed.\n', )
        debug_help = ('', )
        msg = default_msg + msg + debug_help
        res = html.Div(msg)
        res_formatting = measurement_digits('none',
                                            measurement=measurement)
        fig_plot = blank_fig()
        talker_options = none_dropdown
        session_options = none_dropdown
            # )
        return_vals = (
                res,
                res_formatting,
                fig_plot,
                talker_options,
                session_options
                )
    elif measurement == 'diagnostics':
        default_msg = ('Diagnostics object could not be processed.\n', )
        debug_help = ('', )
        msg = default_msg + msg + debug_help
        res = html.Div(msg)
        res_formatting = measurement_digits('none',
                                            measurement=measurement)
        fig_aweight = blank_fig()
        fig_peak = blank_fig()
        fig_fsf= blank_fig()
        return_vals = (
            res,
            res_formatting,
            fig_aweight,
            fig_peak,
            fig_fsf,
            )
    return return_vals


# -*- coding: utf-8 -*-
"""
Created on Tue Oct  5 14:51:35 2021

@author: jkp4
"""
from dash import dcc
from dash import html
import dash

from mcvqoe.hub.eval_app import app
import mcvqoe.hub.eval_shared as eval_shared
# from .eval_app import app

layout = html.Div([
    html.H1('MCV QoE Measurements Processing'),
    
    html.H3('Select a measurement:'),
    
    dcc.Link('Access delay', href='/access',
             style=eval_shared.style_links,
             ),
    html.Br(),
    
    dcc.Link('Intelligibility', href='/intell',
             style=eval_shared.style_links,
             ),
    html.Br(),
    
    dcc.Link('Mouth-to-ear latency', href='/m2e',
             style=eval_shared.style_links,
             ),
    html.Br(),

    dcc.Link('Probability of Successful Delivery', href='/psud',
             style=eval_shared.style_links,
             ),
    html.Br(),
    
    dcc.Link('Transmit Volume Optimization', href='/tvo',
             style=eval_shared.style_links,
             ),
    html.Br(),
    
    html.Hr(),
    dcc.Link('Shutdown', href='shutdown',
             style=eval_shared.style_links,
             ),
    ])


from dash import html
from dash.dependencies import Input, Output
from dash import dash_table

import json
import numpy as np

from mcvqoe.hub.eval_app import app

import mcvqoe.hub.eval_shared as eval_shared
import mcvqoe.accesstime as access


#-----------------------[Begin layout]---------------------------
# TODO: Figure out how to handle correction data (non-default)
measurement = 'access'

layout = eval_shared.layout_template(measurement)

def format_access_results(access_eval,
                        alphas,
                        digits=6,
                        ):
    """
    Format results from access.evaluate object to be in nice HTML.

    Parameters
    ----------
    access_eval : mcvqoe.accesstime.evaluate
        DESCRIPTION.

    Returns
    -------
    children : html.Div
        DESCRIPTION.

    """

    results = []
    for alpha in alphas:
        val, ci = access_eval.eval(alpha)
        intell = alpha * access_eval.fit_data.I0
        res = {
            'Alpha': alpha,
            'Intelligibility': intell,
            'Access Time [seconds]': eval_shared.pretty_numbers(val, digits),
            'Confidence Lower Bound [seconds]': eval_shared.pretty_numbers(ci[0], digits),
            'Confidence Upper Bound [seconds]': eval_shared.pretty_numbers(ci[1], digits),
            }
        results.append(res)
        

    children = html.Div([
        html.P('Uncertainty is expressed with 95% confidence intervals.'),
        dash_table.DataTable(
            columns=[{'name': i, 'id': i} for i in results[0].keys()],
            data=results,
            page_action='native',
            page_size=12,
            )
        ],
        style=eval_shared.style_results,
        )
    return children


# --------------[Callback functions (order matters here!)]--------------------
@app.callback(
    Output(f'{measurement}-measurement-results', 'children'),
    Output(f'{measurement}-measurement-formatting', 'children'),
    Output(f'{measurement}-plot', 'figure'),
    Output(f'{measurement}-intell', 'figure'),
    Output(f'{measurement}-talker-select', 'options'),
    Input(f'{measurement}-json-data', 'data'),
    Input(f'{measurement}-talker-select', 'value'),
    Input(f'{measurement}-show-raw', 'value'),
    Input(f'{measurement}-intell-type', 'value'),
    Input(f'{measurement}-sowc', 'value'),
    Input(f'{measurement}-alpha', 'value'),
    Input(f'{measurement}-measurement-digits', 'value'),
    )
def update_plots(jsonified_data, talker_select, show_raw, intell_type, sowc,
                 alphas, meas_digits,):
    """
    Update all plots

    Parameters
    ----------
    jsonified_data : TYPE
        DESCRIPTION.
    thin : TYPE
        DESCRIPTION.
    talker_select : TYPE
        DESCRIPTION.
    session_select : TYPE
        DESCRIPTION.
    x : TYPE
        DESCRIPTION.

    Returns
    -------
    return_vals : TYPE
        DESCRIPTION.

    """
    
    if jsonified_data is not None:
        
        try:
            # Initialize access eval object
            # access_eval = eval_shared.load_json_data(jsonified_data, f'{measurement}')
            json_data = json.loads(jsonified_data)
            if 'error' in json_data:
                error_out = '. '.join(json_data['error'])
                raise RuntimeError(error_out)
            else:
                access_eval = access.evaluate(json_data=jsonified_data)
            # Translate show_raw, talker_select, intell_type options
            if show_raw == 'True':
                show_raw = True
            elif show_raw == 'False':
                show_raw = False
    
            if talker_select == []:
                talker_select = None
                
            if intell_type == 'intelligibility':
                raw_intell = True
            else:
                raw_intell = False
            
            if sowc == 'True':
                fit_type = 'COR'
            else:
                fit_type = 'NoCOR'
            # Make access plot
            fig_plot = access_eval.plot(raw_intell=raw_intell,
                                        talkers=talker_select,
                                        fit_type=fit_type,
                                        )
            # Make intell plot
            fig_intell = access_eval.plot_intell(talkers=talker_select,
                                                 show_raw=show_raw,
                                                 fit_type=fit_type,
                                                 color_palette=eval_shared.plotly_color_palette,
                                                 )
            # Get talker word combos
            filenames = np.unique(access_eval.data['talker_word'])
            # Initialize dropdown options
            talker_options = [{'label': i, 'value': i} for i in filenames]
            
            # Format table results
            if alphas != []:
                res = format_access_results(access_eval,
                                          digits=meas_digits,
                                          alphas=alphas,
                                          )
                res_formatting = eval_shared.measurement_digits('grid', meas_digits,
                                                                measurement=measurement)
            else:
                res = html.Div('Invalid filters selected')
                res_formatting = eval_shared.measurement_digits('none',
                                                                measurement=measurement)
            return_vals = (
                res,
                res_formatting,
                fig_plot,
                fig_intell,
                talker_options,
                )
        except Exception as e:
            print(e)
            return_vals = eval_shared.failed_process(measurement, msg=e.args)
        
    else:
        return_vals = eval_shared.failed_process(measurement, )

    return return_vals


# -*- coding: utf-8 -*-
"""
Created on Wed Oct 27 15:46:41 2021

@author: jkp4
"""
from dash import dcc
from dash import html
from dash.dependencies import Input, Output, State

import base64
import io
import json
import numpy as np
import os
import pandas as pd
import re

from mcvqoe.hub.eval_app import app

import mcvqoe.hub.eval_shared as eval_shared
import mcvqoe.psud as psud



#-----------------------[Begin layout]---------------------------
# TODO: Say something about common thinning fctor if data can't be thined

measurement = 'intell'
layout = eval_shared.layout_template(f'{measurement}')

# --------------[Functions]----------------------------------
def format_intell_results(intell_eval, digits=6):
    """
    Format results from intelligibility.evaluate object to be in nice HTML.

    Parameters
    ----------
    intell_eval : mcvqoe.intelligibility.evaluate
        DESCRIPTION.

    Returns
    -------
    children : html.Div
        DESCRIPTION.

    """
    pretty_mean = eval_shared.pretty_numbers(intell_eval.mean, digits)
    pretty_ci = eval_shared.pretty_numbers(intell_eval.ci, digits)
    children = html.Div([
        html.H6('Mean intelligibility (scale of 0-1)'),
        html.Div(f'{pretty_mean}'),
        html.H6('95% Confidence Interval'),
        html.Div(f'{pretty_ci}')
        ],
        style=eval_shared.style_results,
        # className='six columns',
        )
    return children

# --------------[Callback functions (order matters here!)]--------------------
@app.callback(
    Output(f'{measurement}-measurement-results', 'children'),
    Output(f'{measurement}-measurement-formatting', 'children'),
    Output(f'{measurement}-scatter', 'figure'),
    Output(f'{measurement}-hist', 'figure'),
    Output(f'{measurement}-bar', 'figure'),
    Output(f'{measurement}-talker-select', 'options'),
    Output(f'{measurement}-session-select', 'options'),
    Input(f'{measurement}-json-data', 'data'),
    Input(f'{measurement}-talker-select', 'value'),
    Input(f'{measurement}-session-select', 'value'),
    Input(f'{measurement}-x-axis', 'value'),
    Input(f'{measurement}-measurement-digits', 'value'),
    )
def update_plots(jsonified_data, talker_select, session_select, x, meas_digits):
    """
    Update all plots

    Parameters
    ----------
    jsonified_data : TYPE
        DESCRIPTION.
    thin : TYPE
        DESCRIPTION.
    talker_select : TYPE
        DESCRIPTION.
    session_select : TYPE
        DESCRIPTION.
    x : TYPE
        DESCRIPTION.

    Returns
    -------
    return_vals : TYPE
        DESCRIPTION.

    """
    
    if jsonified_data is not None:
        try:
            json_data = json.loads(jsonified_data)
            if 'error' in json_data:
                error_out = '. '.join(json_data['error'])
                raise RuntimeError(error_out)
            else:
                intell_eval = eval_shared.load_json_data(jsonified_data, f'{measurement}')
            
            # thinned = thin == 'True'
            if x == 'index':
                x = None
            if talker_select == []:
                talker_select = None
            if session_select == []:
                session_select = None
            
            # TODO: Implement these
            fig_scatter = intell_eval.plot(
                x=x,
                talkers=talker_select,
                test_name=session_select,
                color_palette=eval_shared.plotly_color_palette,
                )
            fig_histogram = intell_eval.histogram(
                talkers=talker_select,
                test_name=session_select,
                )
            fig_bar = intell_eval.bar(
                talkers=talker_select,
                test_name=session_select,
                )
            
            
            
            filenames = intell_eval.data['Filename']
            pattern = re.compile(r'([FM]\d)(?:_b\d{1,2}_w\d)')
            talkers = set()
            for fname in filenames:
                res = pattern.search(fname)
                if res is not None:
                    talkers.add(res.groups()[0])
            talkers = sorted(talkers)
            talker_options = [{'label': i, 'value': i} for i in talkers]
            
            sessions = intell_eval.test_names
            session_options = [{'label': i, 'value': i} for i in sessions]
            
            res = format_intell_results(intell_eval, meas_digits)
            res_formatting = eval_shared.measurement_digits('grid', meas_digits,
                                                            measurement=measurement)
            return_vals = (
                res,
                res_formatting,
                fig_scatter,
                fig_histogram,
                fig_bar,
                talker_options,
                session_options
                )
        except Exception as e:
            print(e)
            return_vals = eval_shared.failed_process(measurement, msg=e.args)
    else:
        return_vals = eval_shared.failed_process(measurement, )
        
    return return_vals


# -*- coding: utf-8 -*-
"""
Created on Fri Jan  7 11:42:08 2022

@author: jkp4
"""
from dash import dcc
from dash import html
from dash.dependencies import Input, Output, State

import json
import numpy as np
import os
import pandas as pd
import tempfile 

import mcvqoe.hub.eval_shared as eval_shared

from mcvqoe.hub.eval_app import app

#-----------------------[Begin layout]---------------------------
# TODO: Say something about common thinning fctor if data can't be thined

measurement = 'tvo'
layout = eval_shared.layout_template(measurement)

# --------------[Functions]----------------------------------
def format_tvo_results(tvo_eval, digits=2):
    """
    Format results from tvo.evaluate object to be in nice HTML.

    Parameters
    ----------
    intell_eval : mcvqoe.intelligibility.evaluate
        DESCRIPTION.

    Returns
    -------
    children : html.Div
        DESCRIPTION.

    """
    pretty_opt = eval_shared.pretty_numbers(tvo_eval.optimal.loc[0, 'Optimum [dB]'], digits)
    pretty_low = eval_shared.pretty_numbers(tvo_eval.optimal.loc[0, 'Lower_Interval [dB]'], digits)
    pretty_up = eval_shared.pretty_numbers(tvo_eval.optimal.loc[0, 'Upper_Interval [dB]'], digits)
    
    children = html.Div([
        html.H6('Optimum transmit Volume (dB)'),
        html.Div(f'{pretty_opt}'),
        html.H6('Optimum interval (dB)'),
        html.Div(f'{[pretty_low, pretty_up]}')
        ],
        style=eval_shared.style_results,
        # className='six columns',
        )
    return children

# --------------[Callback functions (order matters here!)]--------------------
@app.callback(
    Output(f'{measurement}-measurement-results', 'children'),
    Output(f'{measurement}-measurement-formatting', 'children'),
    Output(f'{measurement}-plot', 'figure'),
    Output(f'{measurement}-talker-select', 'options'),
    Output(f'{measurement}-session-select', 'options'),
    Input(f'{measurement}-json-data', 'data'),
    Input(f'{measurement}-talker-select', 'value'),
    Input(f'{measurement}-session-select', 'value'),
    Input(f'{measurement}-x-axis', 'value'),
    Input(f'{measurement}-measurement-digits', 'value'),
    )
def update_plots(jsonified_data, talker_select, session_select, x, meas_digits):
    """
    Update all plots

    Parameters
    ----------
    jsonified_data : TYPE
        DESCRIPTION.
    thin : TYPE
        DESCRIPTION.
    talker_select : TYPE
        DESCRIPTION.
    session_select : TYPE
        DESCRIPTION.
    x : TYPE
        DESCRIPTION.

    Returns
    -------
    return_vals : TYPE
        DESCRIPTION.

    """
    
    if jsonified_data is not None:
        try:
            json_data = json.loads(jsonified_data)
            if 'error' in json_data:
                error_out = '. '.join(json_data['error'])
                raise RuntimeError(error_out)
            else:
                tvo_eval = eval_shared.load_json_data(jsonified_data, f'{measurement}')
            
            # thinned = thin == 'True'
            if x == 'index':
                x = None
            if talker_select == []:
                talker_select = None
            if session_select == []:
                session_select = None
            
            # TODO: Implement these
            fig_plot = tvo_eval.plot(
                talkers=talker_select,
                x=x,
                color_palette=eval_shared.plotly_color_palette,
                )
            # fig_plot = eval_shared.blank_fig()
            
            filenames = np.unique(tvo_eval.data['Filename'])
            
            talker_options = [{'label': i, 'value': i} for i in filenames]
            
            sessions = tvo_eval.test_name
            session_options = [{'label': i, 'value': i} for i in sessions]
            
            res = format_tvo_results(tvo_eval, meas_digits)
            res_formatting = eval_shared.measurement_digits('grid', meas_digits,
                                                            measurement=measurement)
            return_vals = (
                res,
                res_formatting,
                fig_plot,
                talker_options,
                session_options
                )
        except Exception as e:
            print(e)
            return_vals = eval_shared.failed_process(measurement, msg=e.args)
    else:
        return_vals = eval_shared.failed_process(measurement, )
        
    return return_vals


'''
common definitions used across the GUI
'''

import os.path

# Directories to save things in. Will have a sub folder for each measurement
save_dir = os.path.join(os.path.expanduser("~"), 'Documents', 'MCV-QoE')
old_save_dir = os.path.join(os.path.expanduser("~"), 'MCV-QoE')


# -*- coding: utf-8 -*-
"""
Created on Tue Oct  5 14:59:05 2021

@author: jkp4
"""
import argparse
import importlib
import json
import urllib

from dash import dcc
from dash import html
from dash.dependencies import Input, Output
from flask import request
from mcvqoe.hub.eval_app import app, server

import mcvqoe.hub.eval_access as access
import mcvqoe.hub.eval_diagnostics as diag
import mcvqoe.hub.eval_intell as intell
import mcvqoe.hub.eval_measurement_select as measurement_select
import mcvqoe.hub.eval_m2e as m2e
import mcvqoe.hub.eval_psud as psud
import mcvqoe.hub.eval_tvo as tvo


app.layout = html.Div([
    dcc.Location(id='url', refresh=False),
    html.Div(id='page-content'),
    ])

@server.route('/shutdown_request', methods=["GET"])
def shutdown_request():
    
    shutdown()
    # As far as I can tell doesn't matter what this returns, just needs to return something that is not None
    return 'shutting down'
    
def shutdown():
    """
    Shutdown the server so users do not have to hit CTRL+C in terminal
    """
    
    func = request.environ.get('werkzeug.server.shutdown')
    if func is None:
        raise RuntimeError('Not running with the Werkzeug Server')
    func()

def format_data(fpaths, cutpoint, measurement):
    """
    Load local data and store as single json file with cutpoints if applicable.

    Parameters
    ----------
    fpaths : str
        Local paths to data.
    cutpoint : bool
        Flag for if cutpoints are applicable for this measurement or not.

    Raises
    ------
    RuntimeError
        DESCRIPTION.

    Returns
    -------
    final_json : str
        JSON string representation of all required data for measurement loading.

    """
    
    # Initialize dictionary for json info
    modules = {'access': 'mcvqoe.accesstime',
               'intell': 'mcvqoe.intelligibility',
               'm2e': 'mcvqoe.mouth2ear',
               'psud': 'mcvqoe.psud',
               'tvo': 'mcvqoe.tvo',
               'diagnostics': 'mcvqoe.diagnostics',
               }
    try:
        eval_obj = importlib.import_module(modules[measurement]).evaluate(fpaths)
        # eval_obj = eval(modules[measurement]).evaluate(fpaths)
        final_json = eval_obj.to_json()
    except Exception as e:
        out_info = {'test_info': fpaths,
                    'error': e.args,
                    }

        final_json = json.dumps(out_info)

    return final_json

def update_page_data(layout, final_json, measurement):
    """Update relevant json-data element and initial data flag"""
    
    if final_json is not None:
        for child in layout.children:
            if hasattr(child, 'id'):
                if child.id == f'{measurement}-json-data':
                    child.data = final_json
                elif child.id == f'{measurement}-initial-data-passed':
                    child.children = 'True'
    else:
        for child in layout.children:
            if hasattr(child, 'id'):
                # Act like no data loaded yet
                if child.id == f'{measurement}-initial-data-passed':
                    child.children = 'False'

@app.callback(Output('page-content', 'children'),
              Input('url', 'pathname'))
def display_page(pathname):
    
    pathparts = pathname.split(';')
    test_type = pathparts[0]
    measurement = test_type[1:]
    
    if len(pathparts) > 1:
        # We have data to load
        data_files_url = pathparts[1:]
        data_files = [urllib.request.url2pathname(x) for x in data_files_url]
        if measurement in ['psud', 'access', 'accesstime']:
            cutpoints = True
        else:
            cutpoints = False
        final_json = format_data(data_files, cutpoints, measurement)
    else:
        final_json = None

    
    if test_type == '/psud':
        layout = psud.layout
        update_page_data(layout, final_json, measurement)
    
    elif test_type in ['/m2e', '/mouth2ear']:
        layout = m2e.layout
        # Update relevant layout children
        update_page_data(layout, final_json, measurement)
                
    elif test_type in ['/intell', '/intelligibility']:
        layout = intell.layout
        update_page_data(layout, final_json, measurement)
        
    elif test_type in ['/access', '/accesstime']:
        layout = access.layout
        update_page_data(layout, final_json, measurement)
    
    elif test_type in ['/tvo', ]:
        layout = tvo.layout
        update_page_data(layout, final_json, measurement)
        
    elif test_type == '/diagnostics':
        layout = diag.layout
        update_page_data(layout, final_json, measurement)
        
    elif test_type in ['/measurement_select', '/']:
        layout = measurement_select.layout
        
    elif test_type == '/shutdown':
        shutdown()
        layout=html.H3('Successfully shutdown MCV QoE Data App')
        
    else:
        layout = '404'
    return layout


def main():
    parser = argparse.ArgumentParser(
        description=__doc__
        )
    parser.add_argument('--port',
                        default='8050',
                        type=str,
                        help='Port for dash server')
    
    parser.add_argument('-d', '--debug',
                        action='store_true',
                        help='Run dash server in debug mode')
    args = parser.parse_args()
    
    app.run_server(debug=args.debug,
                   port=args.port)
if __name__ == '__main__':
    main()


import importlib.resources
import sys
# used to catch errors from tk
import _tkinter

from mcvqoe.simulation import QoEsim
from os import path
from PIL import Image, ImageTk
from .tk_threading import show_error, Abort_by_User
from .tk_threading import SingletonWindow
from _tkinter import TclError
from tkinter import ttk

import mcvqoe.hub.loadandsave as loadandsave
import scipy.stats as sps
import tkinter as tk
import tkinter.filedialog as fdl


PADX = 10
PADY = 10

FONT_SIZE = 10

# find MCV icon and add set as window icon
def add_mcv_icon(win):
    img_ext = 'ico' if sys.platform.startswith('win') else 'png'
    with importlib.resources.path('mcvqoe.hub','MCV-sm.'+img_ext) as icon:
        if icon:
            try:
                if img_ext == 'ico':
                    # set the title- and taskbar icon
                    win.iconbitmap(icon)
                else:
                    logo = tk.PhotoImage(file=icon)
                    win.iconphoto(True, logo)
            except TclError as e:
                print(f'failed to add icon : {icon}, {e}')
        else:
            print('Could not find icon file')


class ScrollableFrame(ttk.Frame):
    """Used to add a scrollbar to frames. see MCVQoEGui.init_frames() for
    details on how to add a scrollbar to a frame.
    """

    def __init__(self, master, **kwargs):
        self.container = ttk.Frame(master)
        self.canvas = tk.Canvas(self.container)
        self.scrollbar = ttk.Scrollbar(
            self.container, orient='vertical', command=self.canvas.yview)

        super().__init__(self.canvas, **kwargs)

        self.bind(
                "<Configure>",
                self._on_resize
                )
        self.canvas.create_window((0, 0), window=self, anchor="nw")
        self.canvas.configure(yscrollcommand=self.scrollbar.set)
        self.canvas.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)
        self.scrollbar.pack(side=tk.RIGHT, fill=tk.Y)

    def _on_resize(self, e):
        self.canvas.configure(
                    scrollregion=self.canvas.bbox("all")
                    )

    def pack(self, *args, **kwargs):
        self.container.pack(*args, **kwargs)

    def grid(self, *args, **kwargs):
        self.container.grid(*args, **kwargs)
    def place(self, *args, **kwargs):
        self.container.place(*args, **kwargs)
    def pack_forget(self):
        self.container.pack_forget()
    def grid_forget(self):
        self.container.grid_forget()

    def scroll(self, event):

        if (hasattr(event, 'num') and event.num == 5) or (hasattr(
                event, 'delta') and event.delta < 0):
            delta = -1
        else:
            delta = 1
        self.canvas.yview_scroll(-delta, 'units')


class TestCfgFrame(ttk.LabelFrame):
    """
    Base class for frames to configure and run a measurement.


    CLASS VARIABLES
    ---------------

    text : str
        The title of the frame


    PARAMETERS
    ----------

    btnvars : TkVarDict
        loads and stores the values in the controls
    """

    text = ''

    def __init__(self, btnvars, *args, **kwargs):

        kwargs['text'] = self.text
        super().__init__(*args, **kwargs)

        # option functions will get and store their values in here
        self.btnvars = btnvars

        # sets what controls will be in this frame
        controls = self.get_controls()

        # initializes controls
        self.controls = {}
        for row in range(len(controls)):
            c = controls[row](master=self, row=row)

            self.controls[c.__class__.__name__] = c

    def update_title(self, title):
        
        self.configure(text = title)

    def get_controls(self) -> iter:
        """
        Called to acquire an ordered list of control classes to construct.

        Classes should be subclasses of LabeledControl

        Subclasses should override this.


        RETURNS
        -------
        control-list : iter
            a list of control classes that go into the frame, in the order they should
            appear
        """
        
        return tuple()


class SubCfgFrame(TestCfgFrame):
    """Base class to make a subframe for controls grouped together.

    It is designed to behave like a control in and of itself, and should be
    included in the get_controls() of its master.

    CLASS VARIABLES
    ---------------

    text : str
        the title of the frame
    """

    text = ''

    def __init__(self, master, row, *args, **kwargs):

        super().__init__(master.btnvars, master, *args, **kwargs)

        self.grid(column=0, row=row, columnspan=4, sticky='NSEW',
                  padx=PADX, pady=PADY)

        master.controls.update(self.controls)

    def get_controls(self) -> iter:
        """see TestCfgFrame.get_controls"""
        return tuple()


class AdvancedConfigGUI(tk.Toplevel, metaclass=SingletonWindow):
    """A Toplevel window containing advanced/other parameters.

    Used as a base class for all advanced windows, as well as
    hardware- and simulation-settings windows.

    ATTRIBUTES
    ----------

    text : str
        the title of the window
    """
    
    text = ''
    use_scrollbar = False

    def __init__(self, master, btnvars, *args, **kwargs):

        super().__init__(master, *args, **kwargs)
        # hide the window
        self.withdraw()
        # as soon as possible (after app starts) show again
        self.after(0, self.deiconify)

        if self.use_scrollbar:
            self.scroll_frame = ScrollableFrame(master=self)
            self.bot_frame = ttk.Frame(master=self)

            self.scroll_frame.pack(anchor=tk.N, fill=tk.BOTH, expand=True,
                                        side=tk.TOP)
            self.bot_frame.pack(anchor=tk.S, fill=tk.X, side=tk.TOP)

            # keeps track of tcl variable traces for later destruction
            # this prevents some errors in the simulation settings window.
            self.scroll_frame.traces_ = []
        else:
            # keeps track of tcl variable traces for later destruction
            # this prevents some errors in the simulation settings window.
            self.traces_ = []

        # sets its title based on class variable 'text'
        self.title(self.text)

        # sets the controls in this window
        control_classes = list(self.get_controls())

        self.btnvars = btnvars

        if self.use_scrollbar:
            self.scroll_frame.btnvars = btnvars
            self.bot_frame.btnvars = btnvars

        # take keyboard focus
        self.focus_force()

        add_mcv_icon(self)

        self.controls = master.controls
        if self.use_scrollbar:
            self.scroll_frame.controls = self.controls

            # what to add the controls to
            container = self.scroll_frame

            # include the OK button to close the window
            c = _advanced_submit(master=self.bot_frame, row=0)
            c.action = self.destroy
            self.controls[c.__class__.__name__] = c
        else:
            # include the OK button to close the window
            control_classes.append(_advanced_submit)

            # what to add the controls to
            container = self

        # initializes controls
        for row in range(len(control_classes)):
            c = control_classes[row](master=container, row=row)

            # stores controls with their keys being their parameter names
            container.controls[c.__class__.__name__] = c

        if self.use_scrollbar:
            # update things so winfo_width returns good values
            self.scroll_frame.canvas.update_idletasks()
            self.bot_frame.update_idletasks()
            self.bot_frame.update()

            # get scroll region of the canvas, this is the size of the things in it
            scroll_region = self.scroll_frame.canvas.cget('scrollregion').split()

            # compute size of things in the canvas
            cwidth = int(scroll_region[2]) - int(scroll_region[0])
            cheight = int(scroll_region[3]) - int(scroll_region[1])

            # get width of scrollbar
            bar_width = self.scroll_frame.scrollbar.winfo_width()

            # get width of the bottom frame
            bot_height = self.bot_frame.winfo_height()

            width = int(cwidth*1.1) + bar_width

            height = cheight + int(bot_height*1.1)

            # get the height of the screen
            sheight = self.winfo_screenheight()

            # limit maximum height to 80% screen height
            height = min(height, int(sheight*0.8))

            self.geometry(f"{width}x{height}")

        # return key closes window
        self.bind('<Return>', lambda *args : self.destroy())

    def get_controls(self):
        pass

    def destroy(self):
        super().destroy()

        if self.use_scrollbar:
            container = self.scroll_frame
        else:
            container = self

        # remove tcl variable traces (prevents errors in simulation settings)
        for var, trace_id in container.traces_:
            var.trace_remove('write', trace_id)


class DescriptionBlock:
    """
    A block of text to go with Labled controls.

    A block of text that is wordwraped that spans 3 grid rows
    """

    text = ''

    do_font_scaling = True

    MCtrl = ttk.Entry
    MCtrlargs = []
    MCtrlkwargs = {}

    # usually the browse button
    RCtrl = None
    RCtrlkwargs = {}

    padx = PADX
    pady = PADY

    no_value = False # if it has no associated instance variable in measure

    def __init__(self, master, row):
        
        self.master = master

        self.description = ttk.Label(master, text=self.text, wraplength=320)

        self.description.grid(
            padx=self.padx, pady=self.pady, column=0, columnspan=3, row=row, sticky='E')

    def destroy(self):

        # destroy all the things
        if hasattr(self,'description') and self.description is not None:
            # make sure widget exists
            if self.description.winfo_exists():
                self.description.grid_forget()
                self.description.destroy()
            # mark as destroyed
            self.description = None


class LabeledControl:
    """
    A row consisting of the following general template:

        MyParameter:   (?)     [value          ]     BUTTON

    NOTE: if you can, use one of the following base-classes instead of
    over-customizing this one:

        LabeledEntry
        LabeledNumber
        LabeledCheckbox
        LabeledSlider
        MultiChoice
        EntryWithButton
        advanced


    Subclasses should redefine any of the class variables, as well as the
        setup() method

    Subclasses should have their names be equal to the internal parameter name

    CLASS VARIABLES
    ---------------

    text : str
        a user-friendly name for the parameter

    do_font_scaling : bool
        whether the middle control has user-modifiable text that can be font-scaled

        set this to false for things like radiobuttons and checkboxes

    MCtrl : tkinter.ttk widget class
        the class of tkinter widget to use for the middle control

        recommended to use classes from the tkinter.ttk module for aesthetic reasons.

    MCtrlargs : list
        positional arguments to be passed to the MCtrl constructor

    MCtrlkwargs : dict
        kw arguments to be passed to the MCtrl constructor

    variable_arg : 'textvariable', 'variable'
        whether the control's value should be pulled from its text.

        change this to 'variable' for controls like sliders, checkboxes, radiobuttons, etc
        that don't have user-modifiable text

    RCtrl : tkinter.ttk widget class
        the right-most control (usually a button)

    RCtrlkwargs : dict
        the kwargs to be passed to the RCtrl constructor

    padx, pady : int
        the control's padding, in pixels

    no_value : bool
        set this to true for controls, such as the advanced button,
        that do not have a value to put into the measure class


    PARAMETERS  (handled internally)
    ----------
    master :
        the window that this will be in

    row : int
        the row that the controls should be gridded in
    """
    
    text = ''

    do_font_scaling = True

    MCtrl = ttk.Entry
    MCtrlargs = []
    MCtrlkwargs = {}

    variable_arg = 'textvariable'

    # usually the browse button
    RCtrl = None
    RCtrlkwargs = {}

    padx = PADX
    pady = PADY

    no_value = False # if it has no associated instance variable in measure

    def setup(self):
        """additional code that can be run during initiation"""
        
        pass

    def __init__(self, master, row):
        
        self.master = master

        self.l_ctrl = ttk.Label(master, text=self.text)

        self.l_ctrl.grid(
            padx=self.padx, pady=self.pady, column=0, row=row, sticky='E')

        MCtrlkwargs = self.MCtrlkwargs.copy()
        MCtrlargs = self.MCtrlargs.copy()
        RCtrlkwargs = self.RCtrlkwargs.copy()

        # get tcl variable
        if self.__class__.__name__ in master.btnvars:
            self.btnvar = master.btnvars[self.__class__.__name__]

        # help button
        if self.__class__.__doc__:
            self.h_ctrl = HelpIcon(master, tooltext=self._get_help())
            self.h_ctrl.grid(
                column=1, row=row, padx=0, pady=self.pady, sticky='NW')

        self.setup()
        
        # some controls require more flexibility, so they don't use self.MCtrl
        if self.MCtrl:
            try:
                btnvar = self.btnvar
            except AttributeError:
                btnvar = None
                raise KeyError(f" The parameter '{self.__class__.__name__}' "+
                               f"from '{self.master.__class__.__name__}' "+
                               'is missing its default value. '+
                               "Make sure it is declared in 'control_list', "+
                               "and that it is an accepted type")

            if self.variable_arg:
                MCtrlkwargs[self.variable_arg] = self.btnvar

            else:
                MCtrlargs.insert(0, self.btnvar)

            if self.do_font_scaling:
                MCtrlkwargs['font'] = (FONT_SIZE,)

            # initialize the control
            self.m_ctrl = self.MCtrl(master, *MCtrlargs, **MCtrlkwargs)
            self.m_ctrl.grid(
                column=2, row=row, padx=self.padx, pady=self.pady, sticky='WE')

        # Right-most control
        if self.RCtrl:
            # add command to button
            if self.RCtrl in (ttk.Button, tk.Button):
                RCtrlkwargs['command'] = self.on_button

            # initialize the control
            self.r_ctrl = self.RCtrl(master, **RCtrlkwargs)

            self.r_ctrl.grid(
                padx=self.padx, pady=self.pady, column=3, row=row, sticky='WE')

    def _get_help(self):

        lst = []
        for line in self.__class__.__doc__.splitlines():
            lst.append(line.strip())

        return '\n'.join(lst)

    def destroy(self):

        # destroy all the things
        if hasattr(self,'l_ctrl') and self.l_ctrl is not None:
            # make sure window exists
            if self.l_ctrl.winfo_exists():
                self.l_ctrl.grid_forget()
                self.l_ctrl.destroy()
            self.l_ctrl = None

        if hasattr(self,'m_ctrl') and self.m_ctrl is not None:
            # make sure window exists
            if self.m_ctrl.winfo_exists():
                self.m_ctrl.grid_forget()
                self.m_ctrl.destroy()
            self.m_ctrl = None

        if hasattr(self,'r_ctrl') and self.r_ctrl is not None:
            # make sure window exists
            if self.r_ctrl.winfo_exists():
                self.r_ctrl.grid_forget()
                self.r_ctrl.destroy()
            self.r_ctrl = None

        if hasattr(self,'h_ctrl') and self.h_ctrl is not None:
            # make sure window exists
            if self.h_ctrl.winfo_exists():
                self.h_ctrl.grid_forget()
                self.h_ctrl.destroy()
            self.h_ctrl = None

    def on_button(self):
        """The function to run when the user presses the button on the right
        of the control.

        Subclasses should override this.
        """
        
        pass


class HelpIcon(ttk.Button):
    """
    Shows the control's __doc__ when you hover over or select this icon
    """
    
    def __init__(self, master, *args, tooltext, **kwargs):
        kwargs['text'] = '?'
        kwargs['style'] = 'McvHelpBtn.TLabel'
        super().__init__(master, *args, **kwargs)

        # bind hover and selection events
        self.bind('<Enter>', self.enter)
        self.bind('<FocusIn>', self.enter)
        self.bind('<Leave>', self.leave)
        self.bind('<FocusOut>', self.leave)

        self.tw = None
        self.tooltext = tooltext

    def enter(self, event):
        """
        shows the HELP tooltip

        Parameters
        ----------
        event : TYPE
            DESCRIPTION.

        Returns
        -------
        None.
        """
        
        if self.tw:
            return
        self.tw = ToolTip(self, self.tooltext)

        # get location of help icon
        x, y, cx, cy = self.bbox("insert")
        rootx = _get_master(self).winfo_rootx()

        # calculate location of tooltip
        x = x + self.winfo_rootx() - self.tw.winfo_width()
        y = y + cy + self.winfo_rooty() + 27

        # ensure tooltip does not fall off left edge of window
        if x < rootx + 20: x = rootx + 20

        # set position of tooltip
        self.tw.wm_geometry("+%d+%d" % (x, y))

        # show tooltip
        self.tw.show()

    def leave(self, event):
        """
        Hides the tooltip

        Parameters
        ----------
        event : TYPE
            DESCRIPTION.

        Returns
        -------
        None.
        """
        
        if not self.tw:
            return
        self.tw.destroy()
        self.tw = None

def _get_master(m, targets = (tk.Tk, tk.Toplevel)):
    """
    given a tkinter widget, goes up the chain of hierarchy until it finds
    a widget in targets, and returns that widget.

    Used to get a widget's parent window, or to get the tkinter main window


    Parameters
    ----------
    m : widget

    targets : tk classes
        The default is (tk.Tk, tk.Toplevel).

    Returns
    -------
    m : instance of targets
    """
    
    while hasattr(m, 'master'):
            if isinstance(m, targets):
                return m
            m = m.master


class ToolTip(tk.Toplevel):
    """
    A simple tool-tip window for showing help info about a parameter
    """
    
    def __init__(self, master, text, style='McvToolTip.TLabel'):
        super().__init__(master)

        try:
            # on windows, add alwaysontop
            self.attributes('-topmost',True)
        except: pass

        lf = ttk.Frame(self, style='McvToolTip.TFrame')
        lf.pack()

        ttk.Label(lf, text=text, style=style).pack(
            padx=PADX, pady=PADY)

        # removes window title-bar, taskbar icon, etc
        self.wm_overrideredirect(1)

        # temporarily hides window to prevent sudden movement
        self.withdraw()

        # solidifies the window's size to ensure correct placement on screen
        self.update_idletasks()

    def show(self):self.deiconify()


class LabeledEntry(LabeledControl):
    """
    A convenient text-entry control

    CLASS VARIABLES
    ---------------

    text : str
        a user-friendly name for the parameter
    """


class LabeledNumber(LabeledControl):
    """A convenient base class for parameters that should be a number.

    Sub-classes should set any of the following class variables for customization

    CLASS VARIABLES
    ---------------

    text : str
        the user-friendly name for the parameter

    min_ : number
        the minimum value (defaults to 0)
    max_ : number
        the maximum value (defaults to very large)

    increment : number
        how much the up- and down- buttons should change the value.
        this does NOT limit the resolution of the value
    """

    RCtrl = None

    MCtrl = ttk.Spinbox

    min_ = 0
    max_ = 2**15 - 1
    increment = 1

    def __init__(self, *args, **kwargs):

        self.MCtrlkwargs = {'increment':self.increment,
                                          'from_':self.min_,
                                          'to':self.max_}
        super().__init__(*args, **kwargs)


class LabeledSlider(LabeledControl):
    """A convenient base class for parameters that should be a percentage

    Base classes should not modify any of the class variables (except 'text');
    Use LabeledNumber for better resolution and range.

    CLASS VARIABLES
    ---------------

    text : str
        the user-friendly name for the parameter
    """

    RCtrl = None
    MCtrl = None
    # slider does not accept font size
    do_font_scaling = False

    def __init__(self, master, row, *args, **kwargs):
        super().__init__(master, row, *args, **kwargs)
        self.txtvar = self._percentage()
        self.on_change()

        self.btnvar.trace_add('write', self.on_change)

        ttk.Label(master, textvariable=self.txtvar).grid(
            column=3, row=row, sticky='W')

        self.m_ctrl = ttk.Scale(master, variable=self.btnvar,
            from_=0, to=1)

        self.m_ctrl.grid(
            column=2, row=row, padx=self.padx, pady=self.pady, sticky='WE')

    def on_change(self, *args, **kwargs):
        # updates the percentage to match the value of the slider
        self.txtvar.set(self.btnvar.get())


    class _percentage(tk.StringVar):
        """Displays a percentage instead of a float"""
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def set(self, value):
            v = float(value) * 100
            s = ''

            for char in str(v):
                if char == '.':
                    break
                s = f'{s}{char}'

            super().set(f'{s}%')


class _MultiChoice_Frame(tk.Frame):
    
    def __init__(self, *args, association, textvariable, **kwargs):

        super().__init__(*args, **kwargs)
        self.controls = []

        # initialize
        for val, text in association.items():
            c = ttk.Radiobutton(self, variable=textvariable, value=val,
                text=text)

            c.pack(fill=tk.X)

            self.controls.append(c)

    def configure(self, *args, **kwargs):
        for ctrl in self.controls:
            ctrl.configure(*args, **kwargs)


class MultiChoice(LabeledControl):
    """ A base class for multiple-choice controls

    sub-classes should change any of the following class variables
    for customization

    CLASS VARIABLES
    ---------------

    text : str
        user-friendly name for the parameter

    association : dict
        associates the internal option with its user-friendly name. i.e.:

            {
            '1loc'   : '1 Location',
            '2loc_tx': '2 Location (transmit)',
            '2loc_rx': '2 Location (receive)'
            }
    """

    association = {}
    do_font_scaling = False
    MCtrl = _MultiChoice_Frame

    def __init__(self, *args, **kwargs):
        self.MCtrlkwargs = self.MCtrlkwargs.copy()
        self.MCtrlkwargs['association'] = self.association

        super().__init__(*args, **kwargs)


class LabeledCheckbox(LabeledControl):
    """
    Base class for parameters that are boolean

    sub-classes should not change any of the class variables (except 'text')

    CLASS VARIABLES
    ---------------

    text : str
        USE MIDDLE_TEXT INSTEAD

    middle_text : str
        user-friendly name for parameter
    """
    
    middle_text = ''

    MCtrl = ttk.Checkbutton
    do_font_scaling = False
    variable_arg = 'variable'

    def __init__(self, *args, **kwargs):

        self.MCtrlkwargs = {'text': self.middle_text}

        super().__init__(*args, **kwargs)


class EntryWithButton(LabeledControl):
    """A base-class for controls that have an entry alongside a button.

    Subclasses should override the on_button() method to determine the button's action.
    Within the method, the parameter's value can be interacted with using:

        current_value = self.btnvar.get()

        self.btnvar.set(new_value)

    the following class variables should be overridden for customization

    CLASS VARIABLES
    ---------------

    text : str
        user-friendly name for the parameter

    button_text : str
        text to go into the button
    """

    button_text = ''

    RCtrl = ttk.Button

    def on_button(self): pass

    def __init__(self, *args, **kwargs):

        self.RCtrlkwargs = self.RCtrlkwargs.copy()
        self.RCtrlkwargs['text'] = self.button_text
        super().__init__(*args, **kwargs)


class advanced(EntryWithButton):
    """Base class for singular-button 'advanced' controls.

    This could theoretically be sub-classed further for other single-button uses.

    Subclasses should modify the following class variables:

    CLASS VARIABLES
    ---------------

    text : str
        USE BUTTON_TEXT INSTEAD

    toplevel : type
        a subclass of tk.Toplevel that will be constructed when the button is pressed

    button_text : str
        the text inside the button
    """
    
    text = ''

    MCtrl = None

    button_text = 'Advanced...'

    toplevel = None

    no_value = True

    def on_button(self):
        self.toplevel(master=self.master, btnvars=self.master.btnvars)


#---------------------------controls------------------------------------------


class audio_files(EntryWithButton):
    """Audio files to use for testing.

    If left blank, all the files in "Audio Folder" are used.
    """

    text = 'Audio File(s):'

    button_text = 'Browse...'

    def on_button(self):

        initpath = self.master.btnvars['audio_path'].get()
        if not initpath or not path.isdir(initpath):

            # load cached folder
            initpath = loadandsave.fdl_cache[
                f'{self.master.__class__.__name__}.audio_path']

        fp = fdl.askopenfilenames(parent=self.master,
                initialdir=initpath,
                filetypes=[('WAV files', '*.wav')])
        if fp:
            # normalize paths (prevents mixing of / and \ on windows)
            fp = [path.normpath(f) for f in fp]

            path_, files = format_audio_files('', fp)
            self.btnvar.set(files)
            self.master.btnvars['audio_path'].set(path_)

            # cache folder
            loadandsave.fdl_cache.put(
                f'{self.master.__class__.__name__}.audio_path',
                 path_,
                 )


class audio_path(EntryWithButton):
    """The source folder containing the audio files."""


    text = 'Audio Folder:'

    button_text = 'Browse Folder'

    def on_button(self):

        initpath = self.btnvar.get()
        if not initpath or not path.isdir(initpath):

            # load cached folder
            initpath = loadandsave.fdl_cache[
                f'{self.master.__class__.__name__}.audio_path']

        fp = fdl.askdirectory(initialdir = initpath)
        if fp:

            fp = path.normpath(fp)

            path_, files = format_audio_files(path_=fp)

            self.master.btnvars['audio_files'].set(files)

            self.btnvar.set(path_)

            loadandsave.fdl_cache.put(
                f'{self.master.__class__.__name__}.audio_path',
                path_,
                )


class Audio_Set(LabeledControl):
    """Default audio set to use for measurement, if applicable."""

    def __init__(self, master, row, *args, **kwargs):

        self.text = 'Audio Set:'
        self.MCtrl = ttk.Menubutton
        self.do_font_scaling = False

        super().__init__(master, row, *args, **kwargs)

        self.menu = tk.Menu(self.m_ctrl, tearoff=False)

        self.m_ctrl.configure(menu=self.menu)


class trials(LabeledNumber):
    """Number of trials to use for test."""

    text = 'Number of Trials:'

    min_ = 1


class iterations(LabeledNumber):
    """Number of tests to run consecutively"""
    
    text = 'Test Iterations'
    
    min_ = 1
    

class outdir(EntryWithButton):
    """Location to store all output files"""

    text='Output Folder:'

    button_text = 'Browse...'

    def on_button(self):
        dirp = fdl.askdirectory(parent=self.master)
        if dirp:
            dirp = path.normpath(dirp)
            self.btnvar.set(dirp)

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)


class overplay(LabeledNumber):
    """The number of seconds to play silence after the audio is complete.
    This allows for all of the audio to be recorded when there is delay
    in the system"""

    text='Overplay Time (sec):'

    increment = 0.01


class ptt_gap(LabeledNumber):
    """Time to pause after completing one trial and starting the next."""

    text = 'Gap Between Trials:'

    MCtrl = ttk.Spinbox

    increment = 0.01


class SaveAudio(MultiChoice):
    """Sets what recorded audio should be saved into the output folder.

    Set to 'Rx audio only' to only save received audio.
    """
    
    text = 'Save Audio:'

    association = {'all_audio' : 'All audio',
                   'rx_only'   : 'Rx audio only',
                   'no_audio'  : 'No audio',
                   }


# --------------------------- the following 3 are all for pause_trials--------

class RadioCheck(SubCfgFrame):

    text = 'Regular Radio Checks'

    def get_controls(self):
        return (
            _limited_trials,
            pause_trials,
            )


class _limited_trials(LabeledCheckbox):
    """When disabled, sets the number of pause_trials to infinite"""

    middle_text = 'Enable Radio Checks'

    def __init__(self, master, row, *args, **kwargs):

        # give itself an actual stored value
        self.btnvar = tk.BooleanVar()

        super().__init__(master, row, *args, **kwargs)

        # set a trace to change the value depending on 'pause_trials' parameter
        self.btnvar.trace_add('write', self.on_button)
        self.master.btnvars['pause_trials'].trace_add('write', self.update)
        self.update()

    def on_button(self, *args, **kwargs):
        
        if self.btnvar.get():
            val = self.previous
        else:
            val = 'inf'

        self.master.btnvars['pause_trials'].set(val)

    def update(self, *args, **kwargs):

        v = self.master.btnvars['pause_trials'].get()

        other = v != 'inf'
        this = self.btnvar.get()

        if other:
            self.previous = v

        if other != this:
            self.btnvar.set(other)


class pause_trials(LabeledNumber):
    """Number of trials to run before pausing to perform a radio check."""

    text = 'Trials between check:'


    min_ = 1


class time_expand(SubCfgFrame):
    
    text = 'Time Expand'


    def __init__(self, master, row, **kwargs):

        super().__init__(master, row, **kwargs)

    def get_controls(self):

        return (
            _time_expand_i,
            _time_expand_f,
            )


class _time_expand_i(LabeledNumber):
    """Length of time, in seconds, of extra
    audio to send BEFORE the keyword to ABC_MRT16. Adding time protects
    against inaccurate M2E latency calculations and misaligned audio."""

    text = 'Expand Before:'

    increment = 0.01


class _time_expand_f(LabeledNumber):
    """Length of time, in seconds, of extra
    audio to send AFTER the keyword to ABC_MRT16. Adding time protects
    against inaccurate M2E latency calculations and misaligned audio."""

    text = 'Expand After:'

    increment = 0.01


class bgnoise_file(EntryWithButton):
    """This is used to read in a noise file to be mixed with the
    test audio. Default is no background noise."""

    text = 'Noise File:'

    button_text = 'Browse...'

    def on_button(self):
        fp = fdl.askopenfilename(parent=self.master,
            initialfile=self.btnvar.get(),
            filetypes=[('WAV files', '*.wav')])
        if fp:
            self.btnvar.set(fp)


class bgnoise_snr(LabeledNumber):
    """Signal to noise ratio for voice vs noise file."""

    text = 'SNR:'
    min_ = -5
    max_ = 120
    increment = 0.1


class ptt_wait(LabeledNumber):
    """The amount of time to wait, in seconds, between pushing the
    push to talk button and starting playback. This allows time
    for access to be granted on the system."""

    text = 'PTT Wait Time:'

    increment = 0.01


class _advanced_submit(advanced):

    # closes the advanced window

    button_text = 'OK'

    def on_button(self):
        if not hasattr(self, 'action'):
            self.master.destroy()
        else:
            self.action()


# advanced groups

class BgNoise(SubCfgFrame):
    text = 'Background Noise'

    def get_controls(self):
        return (bgnoise_file,
                bgnoise_snr)


# ----------------------- Device delay characterization -----------------------


class dev_dly(LabeledNumber):
    """Delay in seconds of the audio path with no communication device
    present."""

    text = 'Device Delay:'

    increment = 0.001

    # add button functionality too.
    RCtrl = ttk.Button
    RCtrlkwargs = {'text': 'Calibrate'}

    def on_button(self):
        if _get_master(self, tk.Tk).is_simulation.get():
            self.btnvar.set("automatic")
        else:
            CharDevDly()


class CharDevDly(tk.Toplevel, metaclass = SingletonWindow):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # hide the window
        self.withdraw()
        # as soon as possible (after app starts) show again
        self.after(0,self.deiconify)

        add_mcv_icon(self)

        self.title('Device Delay Characterization')

        ttk.Label(self, text=
"""It seems like you have not yet done a characterization test. A
characterization test is a Mouth-to-Ear test that is run with the audio
output directly fed into the input, as shown below.
""").pack(fill=tk.X, padx=10, pady=10)

        # open image
        max_width = 600
        max_height = 500

        canvas = tk.Canvas(self,width=max_width, height=max_height)

        with importlib.resources.path('mcvqoe.hub','dev_dly_char_example.png') as name:
            img = Image.open(name)

        img.thumbnail((max_width,max_height),resample=Image.LANCZOS, reducing_gap=None)

        canvas.crestimg = ImageTk.PhotoImage(img)
        canvas.create_image(
            max_width//2, max_height//2 + 10,
            image=canvas.crestimg
        )

        canvas.pack()

        ttk.Label(self, text='Once finished, enter the device delay below')

        ttk.Button(self, text='Continue', command = self.continue_btn
                   ).pack()

        self.finished = False

    def continue_btn(self):

        _get_master(self, tk.Tk).selected_test.set('DevDlyCharFrame')
        self.destroy()


#------------------------ Global settings for hdw/simulation------------------


class _HdwPrototype:
    """Contains the default audioplayer settings.

    Settings must be drawn from here because the real audioplayer might throw a
    RuntimeError on program load if the interface is not found.
    """
    
    radioport = ''
    default_radio = 1
    blocksize=512
    buffersize=20
    overplay=1.0
    timecode_type='IRIGB_timecode'


# HARDWARE SETTINGS WINDOW
class HdwSettings(AdvancedConfigGUI):
    
    text = 'Hardware Settings'

    # used for _restore_defaults control
    prototype = _HdwPrototype

    def get_controls(self):
        return (
            AudioSettings,
            overplay,
            radioport,
            default_radio,
            timecode_type,
            _restore_defaults,
            )


class radioport(LabeledControl):
    """Port to use for radio interface. Defaults to the first
    port where a radio interface is detected"""

    text = 'Radio Port:'


class default_radio(LabeledControl):
    """PTT output number to use for test."""

    text='PTT Number:'
    MCtrl = ttk.Spinbox
    MCtrlkwargs = {'from_':1, 'to':2}


class timecode_type(LabeledControl):
    """type of timecode to use for two location tests"""

    text = 'Timecode Type:'
    MCtrl = ttk.Combobox
    MCtrlkwargs = {'values' : ('IRIGB_timecode','soft_timecode')}


class AudioSettings(SubCfgFrame):

    text = 'Audio Settings'

    def get_controls(self):
        return (blocksize, buffersize)


class blocksize(LabeledControl):
    """Block size for transmitting audio, must be a power of 2"""

    text = 'Block Size:'
    MCtrl = ttk.Spinbox
    MCtrlkwargs = {'from_':1, 'to':2**15 -1}


class buffersize(LabeledControl):
    """Number of blocks used for buffering audio"""

    text='Buffer Size:'
    MCtrl = ttk.Spinbox
    MCtrlkwargs = {'from_':1, 'to':2**15 -1}


class _restore_defaults(LabeledControl):

    MCtrl = None
    RCtrl = ttk.Button

    RCtrlkwargs = {'text' : 'Restore Defaults'}

    def on_button(self):
        """restores the default hardware/simulation settings."""

        from_obj = self.master.prototype()

        for k, tk_var in self.master.btnvars.items():

            if hasattr(from_obj, k):
                tk_var.set(getattr(from_obj, k))


class _SimPrototype(QoEsim):
    def __init__(self):
        QoEsim.__init__(self)
        self._impairment_plugin = ''


# SIMULATION SETTINGS WINDOW
class SimSettings(AdvancedConfigGUI):

    text = 'Simulation Settings'

    use_scrollbar = True

    # used for _restore_defaults button
    prototype = _SimPrototype

    def get_controls(self):
        return (
            channel_tech,
            channel_rate,

            overplay,

            m2e_latency_cfg,
            access_delay_cfg,
            device_delay_cfg,

            rec_snr,
            PTT_sig_freq,
            PTT_sig_amplitude,
            ImpairmentsSelect,
            _restore_defaults,
            )


class channel_tech(LabeledControl):
    """Technology to use for the simulated channel. Channel technologies are
    handled by plugins. The only tech that is available by default is 'clean'.
    """

    def __init__(self, master, row, *args, **kwargs):

        self.text = 'Channel Tech:'
        self.MCtrl = ttk.Menubutton
        self.do_font_scaling = False

        super().__init__(master, row, *args, **kwargs)

        self.menu = tk.Menu(self.m_ctrl, tearoff=False)

        self.m_ctrl.configure(menu=self.menu)

        try:
            rates = QoEsim.get_channel_techs()
        except Exception as e:
            show_error(e)

        else:
            # get channel_techs to use as menu options
            for tech_ in rates:
                self.menu.add_command(label=tech_,
                        command=tk._setit(self.btnvar, tech_))


class channel_rate(LabeledControl):
    """Rate to simulate channel at. Each channel tech handles this differently.
    When set to None the default rate is used."""

    def __init__(self, master, row, *args, **kwargs):

        self.text = 'Channel Rate:'
        self.MCtrl = ttk.Menubutton

        self.do_font_scaling = False

        super().__init__(master, row, *args, **kwargs)

        self.menu = tk.Menu(self.m_ctrl, tearoff=False)
        self.m_ctrl.configure(menu=self.menu)

        # track selection of channel_tech
        id = self.master.btnvars['channel_tech'].trace_add('write', self.update)
        self.master.traces_.append((self.master.btnvars['channel_tech'], id))

        # fill menu with options
        self.update()


    def update(self, *args, **kwargs):
        
        failed = False
        try:
            chan_tech = self.master.btnvars['channel_tech'].get()

            self.menu.delete(0, 'end')

            default, rates = QoEsim.get_channel_rates(chan_tech)

            old = self.btnvar.get()
            if old not in rates:
                self.btnvar.set(default)

            for rate in rates:
                # add a dropdown list option
                self.menu.add_command(label=str(rate),
                            command=tk._setit(self.btnvar, str(rate)))
        except Exception as e:
            show_error(e)
            failed = True

        if failed:
            self.master.btnvars['channel_tech'].set('clean')


class rec_snr(LabeledControl):
    """Signal to noise ratio for audio channel."""
    
    text = 'Channel SNR:'
    MCtrl = ttk.Spinbox
    MCtrlkwargs = {'from_': 0, 'to': 2**15-1, 'increment':1.0}


class PTT_sig_freq(LabeledControl):
    """Frequency of the PTT signal from the play_record method."""
    
    text = 'PTT Signal Frequency:'
    MCtrl = ttk.Spinbox
    MCtrlkwargs = {'from_':0, 'to' : 2**15-1, 'increment':0.1}


class PTT_sig_amplitude(LabeledControl):
    """Amplitude of the PTT signal from the play_record method."""
    
    text = 'PTT Signal Amplitude:'
    MCtrl = ttk.Spinbox
    MCtrlkwargs = {'from_':0, 'to' : 2**15-1, 'increment':0.1}


class ImpairmentsSelect(SubCfgFrame):

    text = 'Impairments'

    def __init__(self, master, *args, **kwargs):

        self.traces_ = []

        super().__init__(master, *args, **kwargs)

    def get_controls(self):
        return (
            PreImpairment,
            PreImpairmentSettings,
            ChannelImpairment,
            ChannelImpairmentSettings,
            PostImpairment,
            PostImpairmentSettings,
            )


class ChannelImpairment(LabeledControl):
    """Impairment to use on channel data."""

    def __init__(self, master, row, *args, **kwargs):

        self.text = 'Channel Impairment:'
        self.MCtrl = ttk.Menubutton

        self.do_font_scaling = False

        super().__init__(master, row, *args, **kwargs)

        self.menu = tk.Menu(self.m_ctrl, tearoff=False)
        self.m_ctrl.configure(menu=self.menu)

        # track selection of channel_tech
        id = self.master.master.btnvars['channel_tech'].trace_add('write', self.update)
        self.master.master.traces_.append((self.master.btnvars['channel_tech'], id))

        # fill menu with options
        self.update()

    def update(self, *args, **kwargs):
        
        failed = False
        
        try:
            chan_tech = self.master.master.btnvars['channel_tech'].get()

            self.menu.delete(0, 'end')

            if chan_tech == 'None':
                failed = True
                
            else:
                # get channel type
                chan_type = QoEsim().get_channel_type(chan_tech)

                # get list of impairments for the channel type
                impairments = QoEsim.get_impairment_names(chan_type)

                old = self.btnvar.get()
                if old not in impairments:
                    self.btnvar.set('None')

                # add None to the list
                self.menu.add_command(label='None',
                                      command=tk._setit(self.btnvar, 'None'))
                for i in impairments:
                    # add a dropdown list option
                    self.menu.add_command(label=i,
                                command=tk._setit(self.btnvar, i))
        except Exception as e:
            show_error(e)
            failed = True

        if failed:
            self.master.btnvars['channel_tech'].set('None')


# BUG: Garbage collection issue in these delay controls
# ---Reproduce---
# * Open Simulation Settings
# * Set an invalid delay (e.g. '100e-')
# * Close Simulation Settings
# * Reopen Simulation Settings
# * Close Simulation Settings
# * Close the GUI
# ---Trace---
# lib\tkinter\__init__.py", line 388, in __del__
#    if self._tk.getboolean(self._tk.call("info", "exists", self._name)):
# RuntimeError: main thread is not in main loop
# ---Hypotheses---
# Somehow when the default delay is an invalid delay we just lose track of the control
# When it gets garbage collected later, its thread is gone and python complains
# ---Notes---
# This affects all the below delay controls (m2e latency, access delay, device delay)


class m2e_latency_cfg(SubCfgFrame):
    """Config to control m2e latency"""

    text = 'Mouth to Ear Latency'

    def get_controls(self):
        return (
            m2e_latency_type,
            m2e_latency,
            m2e_latency_sigma,
            m2e_latency_range,
            )


class DistributionType(LabeledControl):
    """Control to get the type of a distribution"""

    def __init__(self, master, row, *args, **kwargs):

        self.text = 'Type:'
        self.MCtrl = ttk.Menubutton

        self.do_font_scaling = False

        super().__init__(master, row, *args, **kwargs)

        self.menu = tk.Menu(self.m_ctrl, tearoff=False)
        self.m_ctrl.configure(menu=self.menu)

        # fill menu with options
        self.update()

    def update(self, *args, **kwargs):

        for t in ('constant', 'Normal', 'Exponential'):
            # add a drop down list option
            self.menu.add_command(label=t,
                        command=tk._setit(self.btnvar, t))


class m2e_latency_type(DistributionType):
    """Type of mouth to ear latency"""
    
    # nothing else needs to be configured
    pass


class m2e_latency(LabeledControl):
    """Simulated mouth to ear latency for the channel in seconds.
    'minimum' is the minimum latency for the channel technology."""

    text = 'Value:'
    MCtrl = ttk.Spinbox
    MCtrlkwargs = {'from_': 0, 'to': 2**15-1, 'increment':0.0001}


class m2e_latency_sigma(LabeledControl):
    """
    Sigma value for m2e latency.
    
    If Normal distribution it is standard deviation.
    
    If Exponential distribution it is the inverse of the rate parameter.
    """

    text = '\u03C3:'
    MCtrl = ttk.Spinbox
    MCtrlkwargs = {'from_': 0, 'to': 2**15-1, 'increment':0.0001}


class RangeDisplay:

    text = ''

    padx = PADX
    pady = PADY

    def __init__(self, master, row):
        self.master = master

        ttk.Label(master, text=self.text).grid(
            padx=self.padx, pady=self.pady, column=0, row=row, sticky='E')

        self.range_textvar = tk.StringVar()

        # initialize the control
        self.m_ctrl = ttk.Label(master, textvariable=self.range_textvar)
        self.m_ctrl.grid(
            column=2, columnspan=2, row=row, padx=self.padx, pady=self.pady, sticky='WE')

        # track selection of type
        id = self.master.master.btnvars[self.var_base + '_type'].trace_add('write', self.update)
        self.master.master.traces_.append((self.master.btnvars[self.var_base + '_type'], id))

        # track selection of value
        id = self.master.master.btnvars[self.var_base].trace_add('write', self.update)
        self.master.master.traces_.append((self.master.btnvars[self.var_base], id))

        # track selection of sigma
        id = self.master.master.btnvars[self.var_base + '_sigma'].trace_add('write', self.update)
        self.master.master.traces_.append((self.master.btnvars[self.var_base + '_sigma'], id))

    def update_rng(self, val):
        self.range_textvar.set(str(val))

    def update(self, *args):
        try:
            dist_type = self.master.btnvars[self.var_base + '_type'].get()
            if dist_type != 'constant':
                # enable sigma control
                self.master.controls[self.var_base + '_sigma'].m_ctrl.configure(state='!disabled')
                dist_sigma  = self.master.btnvars[self.var_base + '_sigma'].get()
            else:
                # disable sigma control
                self.master.controls[self.var_base + '_sigma'].m_ctrl.configure(state='disabled')
            dly_val  = self.master.btnvars[self.var_base].get()
        except TclError:
            # failed to get values, clear range
            self.update_rng('')
            return

        if dist_type == 'constant':
            self.update_rng(f'{dly_val}')
        elif dist_type in ['Normal', 'Exponential']:

            if dly_val == 'minimum':
                dly_val = 0
                # set to zero
                # TODO : should this be something else?
                self.master.btnvars['m2e_latency'].set(0)

            try:
                # convert values to float
                dly_val = float(dly_val)
                dist_sigma = float(dist_sigma)
                if dist_type == 'Normal':
                    dist = sps.norm(loc=dly_val, scale=dist_sigma)
                elif dist_type == 'Exponential':
                    dist = sps.expon(loc=dly_val, scale=dist_sigma)

                lower = round(dist.ppf(0.025), 4)
                upper = round(dist.ppf(0.975), 4)
                # update range
                self.update_rng(f'from {lower} to {upper} sec')
            except ValueError:
                # ignore value errors (partially entered number)
                # clear range
                self.update_rng('')


class m2e_latency_range(RangeDisplay):
    """display of the range of the mouth to ear latency"""

    text = 'M2E Latency range\n(roughly 95% of values):'

    var_base = 'm2e_latency'

    def __init__(self, master, row, *args, **kwargs):

        super().__init__(master, row, *args, **kwargs)

        self.update()


class access_delay(LabeledControl):
    """Delay between the time that the simulated push to talk button is pushed
    and when audio starts coming through the channel. If the 'ptt_delay'
    method is called before play_record is called, then the time given to
    'ptt_delay' is added to access_delay to get the time when access is
    granted. Otherwise access is granted 'access_delay' seconds after the
    clip starts."""

    text = 'Value:'
    MCtrl = ttk.Spinbox
    MCtrlkwargs = {'from_': 0, 'to': 2**15-1, 'increment':0.001}


class access_delay_cfg(SubCfgFrame):
    """Config to control access delay"""

    text = 'Access Delay'

    def get_controls(self):
        return (
            access_delay_type,
            access_delay,
            access_delay_sigma,
            access_delay_range,
            )


class access_delay_type(DistributionType):
    """Type of access delay"""
    
    # TODO: Add exponential + constant
    pass


class access_delay_sigma(LabeledControl):
    """
    Sigma value for access delay.
    
    If Normal distribution it is standard deviation
    
    If Exponential distribution it is the inverse of the rate parameter.
    """

    text = '\u03C3:'
    MCtrl = ttk.Spinbox
    MCtrlkwargs = {'from_': 0, 'to': 2**15-1, 'increment':0.0001}


class access_delay_range(RangeDisplay):
    """display of the range of access delay"""

    text = 'Access delay range\n(roughly 95% of values):'
    var_base = 'access_delay'

    def __init__(self, master, row, *args, **kwargs):

        super().__init__(master, row, *args, **kwargs)

        self.update()


class device_delay_cfg(SubCfgFrame):
    """Config to control m2e latency"""

    text = 'Device Delay'

    def get_controls(self):
        return (
            device_delay_type,
            device_delay,
            device_delay_sigma,
            device_delay_range,
            )


class device_delay_type(DistributionType):
    """Type of mouth to ear latency"""
    
    # nothing else needs to be configured
    pass


class device_delay_sigma(LabeledControl):
    """
    Sigma value for device latency.
    
    If Normal distribution it is standard deviation.
    
    If Exponential distribution it is the inverse of the rate parameter.
    """

    text = '\u03C3:'
    MCtrl = ttk.Spinbox
    MCtrlkwargs = {'from_': 0, 'to': 2**15-1, 'increment':0.0001}


class device_delay(LabeledControl):
    """Simulated device delay in seconds."""

    text = 'Device Delay:'
    MCtrl = ttk.Spinbox
    MCtrlkwargs = {'from_': 0, 'to': 2**15-1, 'increment':0.0001}


class device_delay_range(RangeDisplay):
    """display of the range of the mouth to ear latency"""

    text = 'Device delay range\n(roughly 95% of values):'
    var_base = 'device_delay'

    def __init__(self, master, row, *args, **kwargs):

        super().__init__(master, row, *args, **kwargs)

        self.update()


class ImpairmentSettings(SubCfgFrame):

    def __init__(self, master, row, *args, **kwargs):

        self.text = ''

        self.impairment=''

        self.row_num = row

        self.traces_ = []

        super().__init__(master, row, *args, **kwargs)

        # track selection of channel impairment
        id = self.master.btnvars[self.impairment_name].trace_add('write', self.update)
        self.master.traces_.append((self.master.btnvars[self.impairment_name], id))

        self.update()

    def get_controls(self):

        # empty list for controls
        controls = []

        if self.impairment and self.impairment != 'None' and self.winfo_exists():
            params = QoEsim.get_impairment_params(self.impairment)

            description = QoEsim.get_impairment_description(self.impairment)


            # get the name of this class
            cls_name = self.__class__.__name__

            descC = type(f'{cls_name}_Desc',(DescriptionBlock,),{
                                                        'text': description,
                                                      }
                        )

            # append description control to list
            controls.append(descC)

            for name,info in params.items():
                # create type name from class and parameter name
                type_name = f'{cls_name}_{name}'

                # check if a value exists
                if type_name not in self.master.btnvars:
                    #add default
                    self.master.btnvars.add_entry(type_name,info.value_type(info.default))

                if info.choice_type in ('range', 'positive'):

                    class_vals = {
                                    'min_' : info.min_val,
                                    'max_' : info.max_val,
                                    'increment' : info.interval,
                                    'text' : name,
                                 }

                    if hasattr(info,'description'):
                        class_vals['__doc__'] = info.description

                    # create a control class
                    cc = type(type_name,(LabeledNumber,),class_vals)
                    # add to list of controls
                    controls.append(cc)
                elif info.choice_type == 'file':

                    # function to run when browse is clicked
                    def brows_files(self):
                        fp = fdl.askopenfilename(parent=self.master,
                                filetypes=info.filetypes)
                        if fp:
                            # normalize paths (prevents mixing of / and \ on windows)
                            fp = path.normpath(fp)

                            self.btnvar.set(fp)

                    class_vals = {
                                    'on_button' : brows_files,
                                    'button_text' : 'Browse...',
                                    'text' : name,
                                 }

                    if hasattr(info,'description'):
                        class_vals['__doc__'] = info.description

                    # create a control class
                    cc = type(type_name,(EntryWithButton,),class_vals)
                    # add to list of controls
                    controls.append(cc)
                else:
                    print(f'unknown choice type {info.choice_type}')

        return tuple(controls)

    def update(self, *args, **kwargs):

        # update impairment name
        self.impairment = self.master.btnvars[self.impairment_name].get()

        cls_name = self.__class__.__name__

        for n,c in self.master.controls.items():
            if n.startswith(cls_name) and n != cls_name:
                c.destroy()

        # sets what controls will be in this frame
        control_classes = self.get_controls()

        if not self.winfo_exists():
            # window is invalid, bail out here
            return

        # remove remaining widgets from grid
        for c in self.grid_slaves():
            c.grid_forget()

        # remove from grid and forget settings
        if self.impairment == 'None':
            self.grid_forget()
        else:
            # add back to grid
            self.grid(column=0, row=self.row_num, columnspan=4, sticky='NSEW',
                  padx=PADX, pady=PADY)
            self.set_name(self.impairment)

        # initializes controls
        controls = {}
        for row in range(len(control_classes)):
            c = control_classes[row](master=self, row=row)

            controls[c.__class__.__name__] = c

        self.master.controls = controls

    def set_name(self,impairment):
        if impairment:
            self.update_title(f'{impairment} Settings')
        #else:
        #    self.update_title('')


class ChannelImpairmentSettings(ImpairmentSettings):

    def __init__(self, master, row, *args, **kwargs):

        self.impairment_name = 'ChannelImpairment'

        super().__init__(master, row, *args, **kwargs)


class PreImpairment(LabeledControl):
    """impairment to use before audio goes into the channel"""

    def __init__(self, master, row, *args, **kwargs):

        self.text = 'Pre Impairment :'
        self.MCtrl = ttk.Menubutton
        self.do_font_scaling = False

        super().__init__(master, row, *args, **kwargs)

        self.menu = tk.Menu(self.m_ctrl, tearoff=False)

        self.m_ctrl.configure(menu=self.menu)

        try:
            impairments = QoEsim.get_impairment_names('audio')
        except Exception as e:
            show_error(e)

        else:
            # add None to the list
            self.menu.add_command(label='None',
                                  command=tk._setit(self.btnvar, 'None'))
            # get channel_techs to use as menu options
            for i in impairments:
                self.menu.add_command(label=i,
                        command=tk._setit(self.btnvar, i))


class PreImpairmentSettings(ImpairmentSettings):

    def __init__(self, master, row, *args, **kwargs):

        self.impairment_name = 'PreImpairment'

        super().__init__(master, row, *args, **kwargs)


class PostImpairment(LabeledControl):
    """impairment to use after audio goes through the channel"""

    def __init__(self, master, row, *args, **kwargs):

        self.text = 'Post Impairment :'
        self.MCtrl = ttk.Menubutton
        self.do_font_scaling = False

        super().__init__(master, row, *args, **kwargs)

        self.menu = tk.Menu(self.m_ctrl, tearoff=False)

        self.m_ctrl.configure(menu=self.menu)

        try:
            impairments = QoEsim.get_impairment_names('audio')
        except Exception as e:
            show_error(e)

        else:
            # add None to the list
            self.menu.add_command(label='None',
                                  command=tk._setit(self.btnvar, 'None'))
            # get channel_techs to use as menu options
            for i in impairments:
                self.menu.add_command(label=i,
                        command=tk._setit(self.btnvar, i))


class PostImpairmentSettings(ImpairmentSettings):

    def __init__(self, master, row, *args, **kwargs):

        self.impairment_name = 'PostImpairment'

        super().__init__(master, row, *args, **kwargs)


# ---------------------------------- Misc ------------------------------------


class SignalOverride():

    def sig_handler(self, *args, **kwargs):
        # override signal's ability to close the application

        raise Abort_by_User()


def format_audio_files(path_= '', files=[]):
    """
    Parameters
    ----------
    path_ : str, optional
        a path to append all files to. The default is ''.
    files : list of str, optional
        path to each file. The default is [].

    Returns
    -------
    newpath : str
        the absolute path containing all the files.
    newfiles : list of str
        the paths relative to newpath containing the files.
    """

    if not files:
        return path_, ['<entire audio folder>']

    if len(files) == 1:
        f = path.join(path_, files[0])

        if path.isfile(f):

            newpath, newfile = path.split(f)

            return newpath, [newfile]

        elif path.isdir(f):

            return f, ['<entire audio folder>']
        else:

            return '', []

    newpath = path.commonpath([path.join(path_, f) for f in files])

    newfiles = [path.relpath(f, newpath) for f in files]

    return newpath, newfiles


class test(MultiChoice):
    """Test to perform. Options are: 1 Location,
    2 Location transmit, and 2 Location receive."""

    text = 'Location Type:'

    association = {
            '1loc'   : '1 Location',
            '2loc_tx': '2 Location (transmit)',
            '2loc_rx': '2 Location (receive)'
            }



# -*- coding: utf-8 -*-
"""
Created on Tue Oct  5 14:02:51 2021

@author: jkp4
"""
import dash
import flask

external_stylesheets = ['https://codepen.io/chriddyp/pen/bWLwgP.css']
server = flask.Flask(__name__)
app = dash.Dash(__name__,
                server=server,
                suppress_callback_exceptions=True,
                external_stylesheets=external_stylesheets)
# server = app.server



# -*- coding: utf-8 -*-
"""
Created on Tue Jun  1 15:29:51 2021

@author: MkZee
"""

import os
from appdirs import user_data_dir as udd
import json
import tkinter as tk
import _tkinter
import atexit
from .tk_threading import InvalidParameter

import numpy as np


class TkVarDict(dict):
    """A dict of tk variables with initial values given

    Used to save, load, and get states of tkinter widgets
    """

    def __init__(self, **initials):
        for k, v in initials.items():
            self.add_entry(k, v)

    def add_entry(self, key, value, var_type=None):
        if var_type is None:
            if type(value) == bool:
                var_type = tk.BooleanVar
            elif type(value) == str:
                var_type = tk.StringVar
            
            elif np.issubdtype(type(value), np.floating):
                var_type = tk.DoubleVar
            elif type(value) == int:
                var_type = tk.IntVar
            elif type(value) == list:
                var_type = CommaSepList
            elif value is None:
                return
            else:
                print(key, 'has an invalid value. Ignoring.')
                return
        
        self[key] = var_type(value=value)
        self[key].trace_add('write', self._on_change_pre)

        return self[key]

    def set(self, dict_):
        """sets the values of the vars to the values given

        Parameters
        ----------
        dict_ : dict
        """

        for k, v in dict_.items():
            if k in self:
                self[k].set(v)

    def get(self):
        """converts the values back into a dict

        Returns
        -------
        dict_ : dict
        """

        dict_ = {}
        for k, v in self.items():
            try:
                dict_[k] = v.get()
            except _tkinter.TclError:

                message = {
                    tk.BooleanVar: 'Must be either True or False',
                    tk.StringVar : 'Must be a string',
                    tk.DoubleVar : 'Must be a number',
                    tk.IntVar    : 'Must be a whole number',
                    CommaSepList : 'Must be a comma-separated list of strings',

                    }[type(v)]

                raise InvalidParameter(k, message) from None

        return dict_

    def on_change(self):
        # should be set by each instance
        pass

    def _on_change_pre(self, *args, **kwargs):
        self.on_change()


class Config(dict):
    """loads and saves a json file in the user's appdir folder"""

    appinf = {'appname': 'mcvqoe', 'appauthor': 'NIST'}

    def __init__(self, filename, /, **kwargs):
        super().__init__(**kwargs)

        self.filename = filename
        folder = udd(self.appinf['appname'], self.appinf['appauthor'])
        self.filepath = os.path.join(folder, filename)

        os.makedirs(folder, exist_ok=True)

    def dump(self):

        with open(self.filepath, 'w') as fp:
            json.dump(fp=fp, obj=self, indent=1)

    def load(self):
        self.clear()

        with open(self.filepath, 'r+') as fp:
            for k, v in json.load(fp=fp).items():
                self[k] = v
        return self


# -------------------------------- Variable types----------------------------

class CommaSepList(tk.StringVar):
    """used for audio_files"""

    def __init__(self, master=None, value=[], name=None):
        super().__init__(master, ', '.join(value), name)

    def get(self):
        return [x.strip() for x in super().get().split(',')]

    def set(self, value):
        super().set(', '.join(value))


class Vec1Or2Var:
    """used for controls that accept a 1 - or - 2 length vector"""

    def __init__(self, value=[]):
        v = value.copy()

        self.type_ = type(v[0])

        self.zero = tk.StringVar(value=v[0])

        try:
            self.one = tk.StringVar(value=v[1])
        except IndexError:
            self.one = tk.StringVar(value='<default>')

    def trace_add(self, *args, **kwargs):

        self.zero.trace_add(*args, **kwargs)
        self.one.trace_add(*args, **kwargs)

    def get(self):
        lst = [self.type_(self.zero.get())]

        one = self.one.get()
        if one != '<default>':
            lst.append(self.type_(one))

        return lst

    def set(self, v):

        self.zero.set(v[0])

        try:
            self.one.set(v[1])
        except IndexError:
            self.one.set('<default>')


class IntOrInf_OLD(tk.StringVar):

    def __init__(self, master=None, value=0, name=None):

        try:
            val = str(int(value))
        except OverflowError:
            val = 'inf'
        super().__init__(master, val, name)

    def set(self, value):
        try:
            val = str(int(value))
        except OverflowError:
            val = 'inf'
        super().set(val)

    def get(self):
        # return int_or_inf(super().get())
        pass


# ------------------------------------- caches --------------------------------

class BaseCache(Config):

    default = None

    def __init__(self, filename, *args, **kwargs):
        super().__init__(filename, *args, **kwargs)

        try:
            self.load()
        except FileNotFoundError:
            pass

        atexit.register(self.dump)

    def __getitem__(self, k):

        try:
            return super().__getitem__(k)
        except KeyError:
            return self.default

class FdlCache(BaseCache):

    def put(self, key, filepath):
        if os.path.isdir(filepath):
            self[key] = filepath
        elif os.path.isfile(filepath):
            self[key] = os.path.split(filepath)[0]

misc_cache = BaseCache('cache.json')
fdl_cache = FdlCache('file_dialog_cache.json')
dim_cache = BaseCache('window_cache.json')
hardware_settings = BaseCache('hardware_settings.json')
sync_settings = BaseCache('sync_settings.json')



# -*- coding: utf-8 -*-
"""
Created on Fri Dec 10 08:20:53 2021

@author: cjg2
"""
from dash import dash_table
from dash import html
from dash.dependencies import Input, Output

import json

from mcvqoe.hub.eval_app import app

import mcvqoe.hub.eval_shared as eval_shared

#-----------------------[Begin layout]---------------------------

measurement = 'diagnostics'
layout = eval_shared.layout_template(measurement)

def format_diagnostic_results(diag_eval, digits=4):
    """
    Format results from Test_Stats to be in HTML.

    Parameters
    ----------
    diagnostics_eval : 
        DESCRIPTION.

    Returns
    -------
    children : html.Div
        DESCRIPTION.
    """
    df = diag_eval.data
    df.insert(0, 'Trial', df.index + 1)
    
    flagged = df.loc[(df['AW_flag'] == 1) | (df['FSF_flag'] == 1) | (df['Clip_flag'] == 1)]
    
    results = flagged.to_dict(orient='records')
    if len(results) == 0:
        children = [
            html.H3('Flagged Trials'),
            html.P('No flagged trials detected'),
            ]
        
    else:
        children = [
            html.H3('Flagged Trials'),
            html.Div([
            dash_table.DataTable(
                columns=[{'name': i, 'id': i} for i in results[0].keys()],
                data=results,
                page_action='native',
                page_size=12,
                )
            ],
            style=eval_shared.style_results,
            )
            ]
   
 
    return children
    
@app.callback(
    Output(f'{measurement}-measurement-results', 'children'),
    Output(f'{measurement}-measurement-formatting', 'children'),
    Output(f'{measurement}-aweight', 'figure'),
    Output(f'{measurement}-peak', 'figure'),
    Output(f'{measurement}-fsf', 'figure'),
    Input(f'{measurement}-json-data', 'data'),
    Input(f'{measurement}-measurement-digits', 'value'),
    )
def update_plots(jsonified_data,
                 meas_digits,):
    """
    Update all output and plots

    Parameters
    ----------
    jsonified_data : TYPE
        DESCRIPTION.

    Returns
    -------
    return_vals : TYPE
        DESCRIPTION.

    """
    
    if jsonified_data is not None:
        try:
            json_data = json.loads(jsonified_data)
            if 'error' in json_data:
                error_out = '. '.join(json_data['error'])
                raise RuntimeError(error_out)
            else:
                diag_eval = eval_shared.load_json_data(jsonified_data, f'{measurement}')

            fig_aweight = diag_eval.aw_plot()
            fig_peak = diag_eval.peak_dbfs_plot()
            fig_fsf = diag_eval.fsf_plot()
            
            res = format_diagnostic_results(diag_eval, digits=meas_digits)
            res_formatting = eval_shared.measurement_digits('grid', meas_digits,
                                                            measurement=measurement)
            
            return_vals = (
                res,
                res_formatting,
                fig_aweight,
                fig_peak,
                fig_fsf,
                )
        except Exception as e:
            print(e)
            return_vals = eval_shared.failed_process(measurement, msg=e.args)
        
    else:
        return_vals = eval_shared.failed_process(measurement, )
    return return_vals


# -*- coding: utf-8 -*-
"""
Created on Tue Oct 12 12:26:08 2021

@author: jkp4
"""
from dash import dcc
from dash import html
from dash.dependencies import Input, Output, State
from dash import dash_table
import base64
import io
import itertools
import json
import numpy as np
import os
import pandas as pd
import re

from mcvqoe.hub.eval_app import app

import mcvqoe.hub.eval_shared as eval_shared
import mcvqoe.psud as psud



#-----------------------[Begin layout]---------------------------
# TODO: Say something about common thinning fctor if data can't be thined

measurement = 'psud'
layout = eval_shared.layout_template(measurement)

# TODO: PSuD needs radio check buttons for:
    # AMI vs EWC
    # Message Length
    # Intelligibility threshold
    # Display results in a table
# --------------[Functions]----------------------------------
def format_psud_results(psud_eval,
                        thresholds,
                        message_lengths,
                        methods,
                        digits=6,
                        ):
    """
    Format results from psud.evaluate object to be in nice HTML.

    Parameters
    ----------
    psud_eval : mcvqoe.psud.evaluate
        DESCRIPTION.

    Returns
    -------
    children : html.Div
        DESCRIPTION.

    """
    # TODO: Make these controllable at top
    # thresholds = [0.5, 0.7]
    # message_lengths = [1, 3, 5, 10]
    # methods = ['EWC', 'AMI']

    results = []
    for method, thresh, msg_len, in itertools.product(methods, thresholds, message_lengths):
        val, ci = psud_eval.eval(thresh, msg_len, method=method)
        res = {
            'Method': method,
            'Intelligibility Threshold': thresh,
            'Message Length (s)': msg_len,
            'PSuD': eval_shared.pretty_numbers(val, digits),
            'Confidence Lower Bound': eval_shared.pretty_numbers(ci[0], digits),
            'Confidence Upper Bound': eval_shared.pretty_numbers(ci[1], digits),
            }
        results.append(res)
        
    # df_res = pd.DataFrame(results)
        
    # psud_v, psud_ci = psud_eval.eval(0.5, 3)
    # pretty_mean = eval_shared.pretty_numbers(psud_v, digits)
    # pretty_ci = eval_shared.pretty_numbers(psud_ci, digits)
    children = html.Div([
        html.P('Uncertainty is expressed with 95% confidence intervals.'),
        dash_table.DataTable(
            columns=[{'name': i, 'id': i} for i in results[0].keys()],
            data=results,
            page_action='native',
            page_size=12,
            )
        # html.H6('Probability of successful delivery (scale of 0-1)'),
        # html.Div(f'{pretty_mean}'),
        # html.H6('95% Confidence Interval'),
        # html.Div(f'{pretty_ci}')
        ],
        style=eval_shared.style_results,
        # className='six columns',
        )
    return children

# --------------[Callback functions (order matters here!)]--------------------
@app.callback(
    Output(f'{measurement}-measurement-results', 'children'),
    Output(f'{measurement}-measurement-formatting', 'children'),
    Output(f'{measurement}-plot', 'figure'),
    Output(f'{measurement}-scatter', 'figure'),
    Output(f'{measurement}-hist', 'figure'),
    Output(f'{measurement}-talker-select', 'options'),
    Output(f'{measurement}-session-select', 'options'),
    Input(f'{measurement}-json-data', 'data'),
    Input(f'{measurement}-talker-select', 'value'),
    Input(f'{measurement}-session-select', 'value'),
    Input(f'{measurement}-x-axis', 'value'),
    Input(f'{measurement}-intell-type', 'value'),
    Input(f'{measurement}-measurement-digits', 'value'),
    Input(f'{measurement}-method', 'value'),
    Input(f'{measurement}-intelligibility-threshold', 'value'),
    Input(f'{measurement}-message-length', 'value'),
    )
def update_plots(jsonified_data, talker_select, session_select, x, intell_type,
                 meas_digits, method, threshold, message_length):
    """
    Update all plots

    Parameters
    ----------
    jsonified_data : TYPE
        DESCRIPTION.
    thin : TYPE
        DESCRIPTION.
    talker_select : TYPE
        DESCRIPTION.
    session_select : TYPE
        DESCRIPTION.
    x : TYPE
        DESCRIPTION.

    Returns
    -------
    return_vals : TYPE
        DESCRIPTION.

    """
    
    if jsonified_data is not None:
        try:
            json_data = json.loads(jsonified_data)
            if 'error' in json_data:
                error_out = '. '.join(json_data['error'])
                raise RuntimeError(error_out)
            else:
                psud_eval = eval_shared.load_json_data(jsonified_data, f'{measurement}')
            
            # thinned = thin == 'True'
            if x == 'index':
                x = None
            if talker_select == []:
                talker_select = None
            if session_select == []:
                session_select = None

            fig_plot = psud_eval.plot(methods=method,
                                      thresholds=threshold,
                                      color_palette=eval_shared.plotly_color_palette,
                                      )
            fig_scatter = psud_eval.plot_intelligibility(
                x=x,
                data=intell_type,
                talkers=talker_select,
                test_name=session_select,
                color_palette=eval_shared.plotly_color_palette,
                )
            fig_histogram = psud_eval.histogram(color_palette=eval_shared.plotly_color_palette,)
            
            filenames = psud_eval.data['Filename']
            pattern = re.compile(r'([FM]\d)(?:_n\d+_s\d+_c\d+)')
            talkers = set()
            for fname in filenames:
                res = pattern.search(fname)
                if res is not None:
                    talkers.add(res.groups()[0])
            talkers = sorted(talkers)
            talker_options = [{'label': i, 'value': i} for i in talkers]
            
            sessions = np.unique(psud_eval.data['name'])
            session_options = [{'label': i, 'value': i} for i in sessions]            
            
            if method != [] and threshold != [] and message_length != []:
                res = format_psud_results(psud_eval,
                                          digits=meas_digits,
                                          methods=method,
                                          thresholds=threshold,
                                          message_lengths=message_length,
                                          )
                res_formatting = eval_shared.measurement_digits('grid', meas_digits,
                                                                measurement=measurement)
            else:
                res = html.Div('Invalid filters selected')
                res_formatting = eval_shared.measurement_digits('none',
                                                                measurement=measurement)
            return_vals = (
                res,
                res_formatting,
                fig_plot,
                fig_scatter,
                fig_histogram,
                talker_options,
                session_options
                )
        except Exception as e:
            print(e)
            return_vals = eval_shared.failed_process(measurement, msg=e.args)
        
    else:
        return_vals = eval_shared.failed_process(measurement, )
    return return_vals


# -*- coding: utf-8 -*-
"""
Created on Thu Jun 17 14:06:49 2021

@author: MkZee
"""
import os
import mcvqoe.psud as psud
import tkinter as tk
import tkinter.filedialog as fdl

from .shared import LabeledSlider, TestCfgFrame, AdvancedConfigGUI, SignalOverride
from .shared import Audio_Set, iterations
from .shared import format_audio_files, test
from .shared import advanced as shared_advanced

#----------------------------controls-----------------------------------------

from .shared import trials, audio_files, audio_path, outdir
from .shared import ptt_wait, ptt_gap, time_expand
from .shared import SaveAudio, MultiChoice


class m2e_min_corr(LabeledSlider):
    """Minimum correlation value for acceptable mouth 2 ear measurement"""
    text = 'Min Corr. for Success:'

class audio_set(Audio_Set):
    """Included PSuD Audio set to use."""
    text = "Audio Set"
    default_sets, default_path = psud.measure.included_audio_sets()

    def __init__(self, master, row, *args, **kwargs):
        super().__init__(master, row, *args, **kwargs)
        # Grab included audio sets and paths from psud measure class
        sets, audio_path = psud.measure.included_audio_sets()
        # Set default to first set
        self.btnvar.set(self.default_sets[0])
        # Add all sets to drop down
        for aset in self.default_sets:
            self.menu.add_command(
                label=aset,
                command=lambda value=aset: self.update_audio_selection(value)
                )
        # Disable number of trials button

    def update_audio_selection(self, aset):
        """Update audio files and audio paths based on audio set selection."""
        # update audio set button
        self.btnvar.set(aset)
        # Get path to audio set
        path = os.path.join(self.default_path, aset)
        path_, files = format_audio_files(path_=path, files=[])
        # Update audio files and audio paths
        self.master.btnvars['psud_audio_files'].set(files)
        self.master.btnvars['psud_audio_path'].set(path_)
        # Update true controls
        self.master.btnvars['audio_files'].set(files)
        self.master.btnvars['audio_path'].set(path_)
        # Disable Number of trials button
        self.master.controls['psud_trials'].m_ctrl['state'] = 'disable'

class psud_audio_files(audio_files):
    __doc__ = audio_files.__doc__

    def __init__(self, master, row, *args, **kwargs):
        super().__init__(master, row, *args, **kwargs)

    def on_button(self):
        super().on_button()
        # Set psud_audio path to updated path
        new_path = self.master.btnvars['audio_path'].get()
        self.master.btnvars['psud_audio_path'].set(new_path)
        # Update real control for audio files
        self.master.btnvars['audio_files'].set(self.btnvar.get())
        # Enable trials button
        self.master.controls['psud_trials'].m_ctrl['state'] = 'normal'



class psud_audio_path(audio_path):
    __doc__ = audio_path.__doc__
    def __init__(self, master, row, *args, **kwargs):
        super().__init__(master, row, *args, **kwargs)
    
    def on_button(self):
        super().on_button()
        # Update psud audio files
        new_files = self.master.btnvars['audio_files'].get()
        self.master.btnvars['psud_audio_files'].set(new_files)
        # Update real control for audio path
        self.master.btnvars['audio_path'].set(self.btnvar.get())
        # Disable trials
        self.master.controls['psud_trials'].m_ctrl['state'] = 'disable'



class psud_trials(trials):
    __doc__ = trials.__doc__
    
    def __init__(self, master, row, *args, **kwargs):
        super().__init__(master, row, *args, **kwargs)
        self.m_ctrl['state'] = 'disable'
    
    
# ---------------------- The main config frame --------------------------------

class PSuDFrame(TestCfgFrame):
    
    text = 'Probability of Successful Delivery Test'
    
    def get_controls(self):
        return (
            audio_set,
            psud_audio_files,
            psud_audio_path,
            outdir,
            psud_trials,
            iterations,
            SaveAudio,
            ptt_wait,
            ptt_gap,
            test,
            advanced,
            )
    
class PSuDAdvanced(AdvancedConfigGUI):
    
    text = 'PSuD - Advanced'
    
    def get_controls(self):
        return (
            time_expand,
            m2e_min_corr,
            )

class advanced(shared_advanced):
    toplevel = PSuDAdvanced
    
#-------------------------Running the test------------------------------------

class PSuD_fromGui(SignalOverride, psud.measure):
    pass

class PSuD_Eval_from_GUI(SignalOverride, psud.evaluate):
    
    def param_check(self):
        # future-proofing this param-check override
        if hasattr(super(), 'param_check'):
            super().param_check()


# -*- coding: utf-8 -*-
"""
Created on Fri Jun 11 12:50:45 2021

@author: MkZee
"""

import os

import mcvqoe.accesstime as adly

import tkinter as tk

from .shared import TestCfgFrame, SubCfgFrame, SignalOverride, test
from .shared import LabeledNumber, LabeledCheckbox, EntryWithButton, AdvancedConfigGUI
from .shared import advanced as shared_advanced
from .loadandsave import Vec1Or2Var


#------------------------------controls--------------------------------------

from .shared import audio_files, outdir, ptt_gap, time_expand
from .shared import BgNoise, iterations
from .shared import dev_dly, RadioCheck, audio_path, SaveAudio

      
class auto_stop(LabeledCheckbox):
    """Enable checking for access and stopping the test when it is detected."""
    
    middle_text = 'Enable Auto-Stop'


class stop_rep(LabeledNumber):
    """Number of times that access must be detected in a row before the
    test is completed."""
    
    text = 'Stop after __ successful:'
    
    min_ = 1
    
      
class ptt_delay(SubCfgFrame):
    text = 'PTT Delay (sec)'
    
    variable_type = Vec1Or2Var
    
    def get_controls(self):
        return (
            _ptt_delay_min,
            _ptt_delay_max
            )


class _ptt_delay_min(LabeledNumber):
    """The smallest ptt_delay"""
    
    text = 'Min Delay Time:'
    
    increment = 0.01
    
    
class _ptt_delay_max(LabeledNumber):
    """The largest ptt_delay. By default, this will be set to the end
    of the first word in the clip."""
    
    text = 'Max Delay Time:'
    
    increment = 0.01
    

class ptt_step(LabeledNumber):
    """Time difference in seconds between successive ptt_delays."""
    
    text = 'Time Increase per Step:'
    
    increment = 0.01
    

class ptt_rep(LabeledNumber):
    """Number of times to repeat a given PTT delay value. If auto_stop is
    used, this must be greater than 15."""
    
    text = 'Repeats per Step:'
    

class s_thresh(LabeledNumber):
    """The threshold of A-weight power for P2, in dB, below which a trial
    is considered to have no audio."""
    
    text = 'Min Allowed Volume (dB):'
    
    min_ = -2**15 +1
    max_ = 0


class s_tries(LabeledNumber):
    """Number of times to retry the test before giving up."""
    
    text = 'Retry Attempts:'
    
  
class data_file(EntryWithButton):
    """A temporary datafile to use to restart a test. If this is
    given all other parameters are ignored and the settings of the original
    test are used."""
    
    text = 'Recovery File:'
    
    button_text = 'Browse...'
    
    def on_button(self):
        
        outdir = self.master.btnvars['outdir'].get()
        recovery = os.path.join(outdir,'data', 'recovery')
        
        fp = tk.filedialog.askopenfilename(initialdir=recovery)
        if fp:
            self.btnvar.set(fp)


class AutoStop(SubCfgFrame):
    text = 'Auto-Stop'
    
    def get_controls(self):
        return (
            auto_stop,
            stop_rep,
            )
    
  
class DetectFailure(SubCfgFrame):
    text = 'Detecting Failed Transmission'   
    
    def get_controls(self):
        return (
            s_thresh,
            s_tries,
            )


class zip_audio(LabeledCheckbox):
    """ If checked audio will be zipped after test completes."""

    middle_text = "Zip Audio"
    

class bisect_midpoint(LabeledCheckbox):
    """If checked PTT times will be determined iteratively and will attempt to 
    converge around the PTT time associated with the intelligibility midpoint
    of the intelligibility curve. This will generally result in a much faster test,
    but may be more susceptible to generating an invalid intelligibility curve and
    access delay result in extreme circumstances.
    
    If unchecked PTT times will be uniformly spaced, and are predetermined based on
    other settings. This is the "safest" option in some ways, but generally 
    results in much longer tests.
    """
    text = "Bisect Midpoint"


# ------------------------ The main configuration frame -----------------------


class AccssDFrame(TestCfgFrame):
    text = 'Access Delay Test'
    
    
    def get_controls(self):
        return (
            audio_files,
            audio_path,
            outdir,
            iterations,
            ptt_step,
            ptt_gap,
            SaveAudio,
            RadioCheck,
            dev_dly,
            data_file,
            test,
            advanced,
            )


# -------------------------- The advanced window ------------------------------


class AccDlyAdvanced(AdvancedConfigGUI):
    text = 'Access Delay - Advanced'
    
    def get_controls(self):
        return (
            
            ptt_rep,
            zip_audio,
            ptt_delay,
            
            bisect_midpoint,
            
            AutoStop,
            BgNoise,
            DetectFailure,
            time_expand,
            )
        
    
class advanced(shared_advanced):
    toplevel = AccDlyAdvanced


class Access_fromGui(SignalOverride, adly.measure):
    pass
    

class Access_Eval_from_GUI(SignalOverride, adly.evaluate):
    
    def param_check(self):
        # future-proofing this param-check override
        if hasattr(super(), 'param_check'):
            super().param_check()
